"""
Interface to publish functions to PiCloud, allowing them to be invoked via the REST API

Api keys must be configured before using any functions in this module.
(via scicloudconf.py, scicloud.setkey, or being on PiCloud server)
"""
from __future__ import absolute_import
"""
Copyright (c) 2011 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@piscicloud.com

The scicloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see 
http://www.gnu.org/licenses/lgpl-2.1.html
"""
import sys
import re
import __builtin__

from . import _getscicloud, _getscicloudnetconnection
from .util.zip_packer import Packer
from .util import getargspec

try:
    from json import dumps as json_serialize
    from json import loads as json_deserialize
except ImportError: #If python version < 2.6, we need to use simplejson
    from simplejson import dumps as json_serialize
    from simplejson import loads as json_deserialize

_publish_query = 'rest/register/'
_remove_query = 'rest/deregister/'
_list_query = 'rest/list/'
_info_query = 'rest/info/'
_invoke_query = 'rest/invoke/%s/'
_invoke_map_query = 'rest/invoke/%s/map/'

"""
This module utilizes the scicloud object extensively
The functions can be viewed as instance methods of the Cloud (hence accessing of protected variables)
"""

   

def _low_level_publish(func, label, out_encoding, arg_encoding, params, func_desc):
    scicloud = _getscicloud()
    conn = _getscicloudnetconnection()
        
    os_env_vars = params.pop('os_env_vars', None)
    sfunc, sarg, logprefix, logcnt = scicloud.adapter.scicloud_serialize(func, 
        params['fast_serialization'], 
        [], logprefix='rest.', os_env_vars=os_env_vars)

    #Below is derived from HttpAdapter.job_add
    conn._update_params(params)
    scicloud.adapter.dep_snapshot() #let adapter make any needed calls for dep tracking
    data = Packer()
    data.add(sfunc)
    params['data'] = data.finish()
    params['label'] = label
    params['description'] = func_desc
    params['arg_encoding'] = arg_encoding
    params['out_encoding'] = out_encoding
    resp = conn.send_request(_publish_query, params)
    return resp

def publish(func, label, out_encoding='json', **kwargs):
    """       
    Publish *func* (a callable) to PiCloud so it can be invoked through the PiCloud REST API
    
    The published function will be managed in the future by a unique (URL encoded) *label*. 
    
    *out_encoding* specifies the format that the return value should be in when retrieving the result
    via the REST API. Valid values are "json" for a JSON-encoded object and "raw", where the return value
    must be an str (but can contain any characters).
    
    The return value is the URL which can be HTTP POSTed to to invoke *func*. 
    See http://docs.scivm.com/rest.html for information about PiCloud's REST API    
    
    Certain special *kwargs* associated with scicloud.call can be attached to the periodic jobs: 
        
    * _cores:
        Set number of cores your job will utilize. See http://docs.scivm.com/primer.html#choose-a-core-type/
        In addition to having access to more CPU cores, the amount of RAM available will grow linearly.
        Possible values for ``_cores`` depend on what ``_type`` you choose:
        
        * c1: 1                    
        * c2: 1, 2, 4, 8
        * f2: 1, 2, 4, 8, 16                                    
        * m1: 1, 2, 4, 8
        * s1: 1        
    * _env:
        A string specifying a custom environment you wish to run your job within.
        See environments overview at http://docs.scivm.com/environment.html                
    * _fast_serialization:
        This keyword can be used to speed up serialization, at the cost of some functionality.
        This affects the serialization of the spawned jobs' return value.
        The stored function will always be serialized by the enhanced serializer, with debugging features.
        Possible values keyword are:
                    
        0. default -- use scicloud module's enhanced serialization and debugging info            
        1. no debug -- Disable all debugging features for result            
        2. use cPickle -- Use python's fast serializer, possibly causing PicklingErrors
    * _max_runtime:
        Specify the maximum amount of time (in integer minutes) jobs can run. If the job runs beyond 
        this time, it will be killed.                    
    * _os_env_vars:
        List of operating system environment variables that should be copied to PiCloud from your system
        Alternatively a dictionary mapping the environment variables to the desired values.                                                                            
    * _priority: 
        A positive integer denoting the job's priority. PiCloud tries to run jobs 
        with lower priority numbers before jobs with higher priority numbers.            
    * _profile:
        Set this to True to enable profiling of your code. Profiling information is 
        valuable for debugging, but may slow down your jobs.
    * _restartable:
        In the very rare event of hardware failure, this flag indicates that a spawned 
        job can be restarted if the failure happened in the middle of it.
        By default, this is true. This should be unset if the function has external state
        (e.g. it modifies a database entry)
    * _type:
        Choose the type of core to use, specified as a string:
        
        * c1: 1 compute unit, 300 MB ram, low I/O (default)                    
        * c2: 2.5 compute units, 800 MB ram, medium I/O
        * f2: 5.5 compute units, 3.75 GB ram, high I/O, hyperthreaded core                                    
        * m1: 3.25 compute units, 8 GB ram, high I/O
        * s1: Up to 2 compute units (variable), 300 MB ram, low I/O, 1 IP per core
                       
        See http://www.scivm.com/pricing/ for pricing information
    * _vol:
        A string or list of strings specifying a volume(s) you wish your jobs to have access to.

    """
    
    if not callable(func):
        raise TypeError( 'scicloud.rest.publish first argument (%s) is not callable'  % (str(func) ))        
    
    m = re.match(r'^[A-Z0-9a-z_+-.]+$', label)
    if not m:
        raise TypeError('Label can only consist of valid URI characters (alphanumeric or from set(_+-.$)')
        
    #ASCII label:
    try:
        label = label.decode('ascii').encode('ascii')
    except (UnicodeDecodeError, UnicodeEncodeError): #should not be possible
        raise TypeError('label must be an ASCII string')
    
    try:
        docstring = '' if (func.__doc__ is None) else func.__doc__
        func_desc = (docstring).encode('utf8')
    except (UnicodeDecodeError, UnicodeEncodeError):
        raise TypeError('function docstring must be an UTF8 compatible unicode string')
    
    if not isinstance(out_encoding, str):
        raise TypeError('out_encoding must be an ASCII string')
    
    scicloud = _getscicloud()
    params = scicloud._getJobParameters(func, kwargs, 
                                             ignore=['_label', '_depends_on', '_depends_on_errors'])
    
    #argument specification for error checking and visibility
    argspec = getargspec(func)    
    argspec_serialized = json_serialize(argspec,default=str)
    if len(argspec_serialized) >= 255: #won't fit in db - clear defaults
        argspec[4] = {}
        argspec_serialized = json_serialize(argspec,default=str)        
    params['argspec'] = argspec_serialized
    
    resp = _low_level_publish(func, label, out_encoding, 'raw', params, func_desc)
    
    return resp['uri']

register = publish 

def remove(label):
    """
    Remove a published function from PiCloud
    """

    #ASCII label:
    try:
        label = label.decode('ascii').encode('ascii')
    except (UnicodeDecodeError, UnicodeEncodeError):
        raise TypeError('label must be an ASCII string')
    
    conn = _getscicloudnetconnection()
    
    conn.send_request(_remove_query, {'label': label})   
    
deregister = remove    
   
def list():
    """
    List labels of published functions
    """ 
    #note beware: List is defined as this function
    conn = _getscicloudnetconnection()
    resp = conn.send_request(_list_query, {})
    return resp['labels']

def info(label):
    """
    Retrieve information about a published function specified by *label*
    """
    conn = _getscicloudnetconnection()
    resp = conn.send_request(_info_query, {'label':label})
    del resp['data']
    
    return resp

def _invoke(label, argument_dict, is_map=False):
    """low-level invoking with map ability"""
    # TODO: We actually have to json_serialize this.. somehows
    
    conn = _getscicloudnetconnection()
    
    map_limit = conn.map_job_limit
    map_count = 1
    
    extracted_argdict = {}
    for name, arglist in argument_dict.items():
        extracted = []
        
        if not hasattr(arglist, '__iter__'):
            raise TypeError('%s must map to an iterable' % name)
            
        argiter = iter(arglist)
        
        iterated = False
        for arg in argiter:
            iterated = True
            if not hasattr(arg, 'read'): # encode anything but file objects
                try:
                    arg = json_serialize(arg)
                except (TypeError, UnicodeDecodeError):
                    raise TypeError('%s is not json encodable' % name)
            extracted.append(arg)
        
        if not iterated:
            raise ValueError('%s cannot be bound to an empty list' % name)
        
        extracted_len = len(extracted)
        if map_count == 1:
            if not is_map and extracted_len > 1:
                raise ValueError('%s can only have 1 item when allow_map is False' % name)
            
            if extracted_len > map_limit:
                raise ValueError('%s has %s items. Maximum is %s', name, extracted_len, map_limit)        
                        
            map_count = extracted_len
        elif extracted_len > 1 and extracted_len != map_count:
            raise ValueError('%s has %s items. Expected %s to match with other arguments' % (name, extracted_len, map_count))        
    
        extracted_argdict[name] = extracted    
    
    if is_map:        
        resp = conn.send_request(_invoke_map_query % label, extracted_argdict)        
        jid_res = resp['jids']
        rstart, rend = jid_res.split('-')
        return xrange(int(rstart), int(rend))
            
        
    else:
        resp = conn.send_request(_invoke_query % label, extracted_argdict)        
        return resp['jid']
            

def invoke(label, **kwargs):    
    """
    Explicitly invoke rest function defined by *label* with *kwargs*
    If value of a kwarg is a File-like object, it will be transmitted as binary data
    Otherwise, if value is a python object, json encoded version of it will be transmitted
    
    Returns jid
    
    """
    listified_args = dict((name, [val]) for name, val in kwargs.items())
    return _invoke(label, listified_args, is_map = False)

def invoke_map(label, **kwargs):    
    """
    Explicitly invoke rest function defined by *label* with *kwargs*
    Emulates PiCloud rest semantics:
    
    args is a dictionary of arguments to map.
        key is argument name        
        value is a list of arguments. 1 argument maps to one map job 
        if value has len(1), value will apply to all map jobs 
        All values with len(>1) must have same length
        Values need to be either file-like objects or json-encodable python objects         
        
    Be aware that this is more limiting than regular scicloud.map
    -No more than 500 jobs
    -All of args will be read into memory before transport
    You will need to manually "chunk" maps if your arguments do not fit the above constraints
    """ 
    return _invoke(label, kwargs, is_map = True)



'''
Defines the logic necessary to handle templates and argument substitution for shell exec
'''
import re
from collections import defaultdict

variable_extract = re.compile(r'(?:[^\$\\]|\A){(\w+?)}')
def extract_vars(command_str):
    """Extract variables from a command string"""    
    matches = variable_extract.findall(command_str)
    return list(set(matches))

variable_extract_dup = re.compile(r'([^\$\\]|\A){{(\w+?)}}') # matches vars in duplicate curlies
def generate_command(command_str, var_dct, skip_validate = False):
    """Fill in variables in command_str with ones from var_dct"""
        
    if not skip_validate:
        validate_command_args(command_str, var_dct)
    
    # first duplicate all curlies    
    command_str = command_str.replace('{', '{{')
    command_str = command_str.replace('}', '}}')
    #print command_str
    
    # now un-duplicate template variables    
    command_str = variable_extract_dup.sub('\\1{\\2}', command_str)
    #print command_str
    
    formatted_cmd =  command_str.format(**var_dct)
    # replace escaped items
    formatted_cmd = formatted_cmd.replace('\\{', '{')
    formatted_cmd = formatted_cmd.replace('\\}', '}')
    return formatted_cmd


def validate_command_args(command_str, var_dct):
    command_vars = extract_vars(command_str)
    for var in command_vars:
        if var not in var_dct:
            raise ValueError('Paremeter %s in command "%s" was not bound' % (var, command_str))
    for var in var_dct:
        if var and var not in command_vars:
            raise ValueError('Argument named %s is not defined in command "%s"' % (var, command_str))

def _var_format_error(item):
    return ValueError('%s: Incorrect format. Variables must be formatted as name=value' % item)

def extract_args(arg_list, allow_map = False):
    """Returns dictionary mapping keyword to list of arguments.
    every list should be of length one if allow_map is false
    """
    kwds = {}
    if not arg_list:
        return kwds
    for arg in arg_list:
        
        parts = arg.split('=', 1)
        if len(parts) != 2:
            raise _var_format_error(arg)
        key, value = parts
        if not key or not value:
            raise _var_format_error(arg)
        if key in kwds:
            raise ValueError('key %s is multiply defined' % key)
        if not allow_map:
            kwds[key] = value
        else:
            kwd_values = []
            # split string on non-escaped ','
            buf_str = ''
            while True:
                idx = value.find(',')
                if idx == -1:
                    break
                if value[idx - 1] == '\\': #escaped
                    buf_str = buf_str + value[:idx+1]                    
                else:
                    kwd_values.append(buf_str + value[:idx])
                    buf_str = ''
                value = value[idx+1:]                
            if buf_str or value:
                kwd_values.append(buf_str + value)
            kwds[key] = kwd_values
    return kwds


    
        

if __name__ == '__main__':
    cmdstr = 'base'
    print generate_command(cmdstr, {})
    cmdstr = 'bash {} ${env} {1..2}'
    print generate_command(cmdstr, {})
    cmdstr = '{hello} bash {} ${{env_sub}} {1..2} {bye}'
    print generate_command(cmdstr, {'hello' : 'HELLO',
                                    'env_sub' : 'ENV_SUB',
                                    'bye' : 'BYE'})
    cmdstr = '{hello} bash {} ${{env_sub}} {1..2} \{bye\}'
    print generate_command(cmdstr, {'hello' : 'HELLO',
                                    'env_sub' : 'ENV_SUB'})            


"""
Science VM realtime management.
This module allows the user to manage their realtime cores programmatically.
See http://docs.scim.com/tech_overview.html#scheduling
"""
from __future__ import absolute_import
"""
Copyright (c) 2011 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@piscicloud.com

The scicloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see 
http://www.gnu.org/licenses/lgpl-2.1.html    
"""


import scicloud
import types
from scicloud.util import fix_time_element

import logging, datetime


_request_query = 'realtime/request/'
_release_query = 'realtime/release/'
_list_query = 'realtime/list/'
_change_max_duration_query = 'realtime/change_max_duration/'


"""
Real time requests management
"""
def list(request_id=""):
    """Returns a list of dictionaries describing realtime core requests.
    If *request_id* is specified, only show realtime core request with that request_id
    
    The keys within each returned dictionary are:
    
    * request_id: numeric ID associated with the request 
    * type: Type of computation resource this request grants
    * cores: Number of (type) cores this request grants
    * start_time: Time when real time request was satisfied; None if still pending"""
    
    if request_id != "":
        try:
            int(request_id)
        except ValueError:
            raise TypeError('Optional parameter to list_rt_cores must be a numeric request_id')
    
    conn = scicloud._getscicloudnetconnection()
    rt_list = conn.send_request(_list_query, {'rid': str(request_id)})
    return [fix_time_element(rt,'start_time') for rt in rt_list['requests']]

def request(type, cores, max_duration=None):
    """Request a number of *cores* of a certain compute resource *type*  
    Returns a dictionary describing the newly created realtime request, with the same format
    as the requests returned by list_rt_cores.
    If specified, request will terminate after being active for *max_duration* hours
    """
    
    if max_duration != None:
        if not isinstance(max_duration, (int, long)):
            raise TypeError('Optional parameter max_duration should be an integer value > 0')
        if max_duration <= 0:
            raise TypeError('Optional parameter max_duration should be an integer value > 0')
    
    conn = scicloud._getscicloudnetconnection()
    return fix_time_element(conn.send_request(_request_query, 
                                               {'cores': cores,
                                                'type' : type,
                                                'cap_duration': max_duration if max_duration else 0}), 
                             'start_time')

def release(request_id):
    """Release the realtime core request associated with *request_id*. 
    Request must have been satisfied to terminate."""
    
    try:
        int(request_id)
    except ValueError:
        raise TypeError('release_rt_cores requires a numeric request_id')
    
    conn = scicloud._getscicloudnetconnection()
    conn.send_request(_release_query, {'rid': str(request_id)})

def change_max_duration(request_id, new_max_duration=None):

    try:
        int(request_id)
    except ValueError:
        raise TypeError('release_rt_cores requires a numeric request_id')
    
    if new_max_duration != None:
        if not isinstance(new_max_duration, (int, long)):
            raise TypeError('Optional parameter max_duration should be an integer value > 0')
        if new_max_duration <= 0:
            raise TypeError('Optional parameter max_duration should be an integer value > 0')
    
    conn = scicloud._getscicloudnetconnection()
    
    conn.send_request(_change_max_duration_query, {'rid': str(request_id), 'cap_duration':new_max_duration})




"""Functions that read and write gzipped files.

The user of the file doesn't have to worry about the compression,
but random access is not allowed."""

# based on Andrew Kuchling's minigzip.py distributed with the zlib module

#This is a patched version that allows seek
#It is under the PSF license

import struct, sys, time
import zlib
import __builtin__

__all__ = ["GzipFile","open"]

FTEXT, FHCRC, FEXTRA, FNAME, FCOMMENT = 1, 2, 4, 8, 16

READ, WRITE = 1, 2

def LOWU32(i):
    """Return the low-order 32 bits of an int, as a non-negative int."""
    return i & 0xFFFFFFFFL


def write32u(output, value):
    # The L format writes the bit pattern correctly whether signed
    # or unsigned.
    output.write(struct.pack("<L", value))

def read32(input):
    return struct.unpack("<I", input.read(4))[0]

def open(filename, mode="rb", compresslevel=9):
    """Shorthand for GzipFile(filename, mode, compresslevel).

    The filename argument is required; mode defaults to 'rb'
    and compresslevel defaults to 9.

    """
    return GzipFile(filename, mode, compresslevel)

class PaddedFile:
    """readonly file object that prepends string the content of f"""
    def __init__(self, string, f):
        self._string = string
        self._length = len(string)
        self._file = f
        self._read = 0

    def read(self, size):
        if self._read is None:
            return self._file.read(size)
        if self._read + size <= self._length:
            read = self._read
            self._read += size
            return self._string[read:self._read]
        else:
            read = self._read
            self._read = None
            return self._string[read:] + \
                   self._file.read(size-self._length+read)

    def unused(self):
        if self._read is None:
            return ''
        return self._string[self._read:]
        

class GzipFile:
    """The GzipFile class simulates most of the methods of a file object with
    the exception of the readinto() and truncate() methods.

    """

    myfileobj = None

    def __init__(self, filename=None, mode=None,
                 compresslevel=9, fileobj=None):
        """Constructor for the GzipFile class.

        At least one of fileobj and filename must be given a
        non-trivial value.

        The new class instance is based on fileobj, which can be a regular
        file, a StringIO object, or any other object which simulates a file.
        It defaults to None, in which case filename is opened to provide
        a file object.

        When fileobj is not None, the filename argument is only used to be
        included in the gzip file header, which may includes the original
        filename of the uncompressed file.  It defaults to the filename of
        fileobj, if discernible; otherwise, it defaults to the empty string,
        and in this case the original filename is not included in the header.

        The mode argument can be any of 'r', 'rb', 'a', 'ab', 'w', or 'wb',
        depending on whether the file will be read or written.  The default
        is the mode of fileobj if discernible; otherwise, the default is 'rb'.
        Be aware that only the 'rb', 'ab', and 'wb' values should be used
        for cross-platform portability.

        The compresslevel argument is an integer from 1 to 9 controlling the
        level of compression; 1 is fastest and produces the least compression,
        and 9 is slowest and produces the most compression.  The default is 9.

        """

        # guarantee the file is opened in binary mode on platforms
        # that care about that sort of thing
        if mode and 'b' not in mode:
            mode += 'b'
        if fileobj is None:
            fileobj = self.myfileobj = __builtin__.open(filename, mode or 'rb')
        if filename is None:
            if hasattr(fileobj, 'name'): filename = fileobj.name
            else: filename = ''
        if mode is None:
            if hasattr(fileobj, 'mode'): mode = fileobj.mode
            else: mode = 'rb'

        if mode[0:1] == 'r':
            self.mode = READ
            # Set flag indicating start of a new member
            self._new_member = True

            self.decompobj = zlib.decompressobj(-zlib.MAX_WBITS)
            self.crcval = zlib.crc32("")
            self.buffer = []   # List of data blocks
            self.bufferlen = 0
            self.pos = 0       # Offset of next data to read from buffer[0]
            self.name = filename
            # Starts small, scales exponentially
            self.min_readsize = 100

        elif mode[0:1] == 'w' or mode[0:1] == 'a':
            self.mode = WRITE
            self._init_write(filename)
            self.compress = zlib.compressobj(compresslevel,
                                             zlib.DEFLATED,
                                             -zlib.MAX_WBITS,
                                             zlib.DEF_MEM_LEVEL,
                                             0)
        else:
            raise IOError, "Mode " + mode + " not supported"

        self.size = 0
        self.fileobj = fileobj
        self.offset = 0

        if self.mode == WRITE:
            self._write_gzip_header()

    @property
    def filename(self):
        import warnings
        warnings.warn("use the name attribute", DeprecationWarning, 2)
        if self.mode == WRITE and self.name[-3:] != ".gz":
            return self.name + ".gz"
        return self.name

    def __repr__(self):
        s = repr(self.fileobj)
        return '<gzip ' + s[1:-1] + ' ' + hex(id(self)) + '>'

    def _init_write(self, filename):
        self.name = filename
        self.crc = zlib.crc32("") & 0xffffffffL
        self.writebuf = []
        self.bufsize = 0

    def _write_gzip_header(self):
        self.fileobj.write('\037\213')             # magic header
        self.fileobj.write('\010')                 # compression method
        fname = self.name
        if fname.endswith(".gz"):
            fname = fname[:-3]
        flags = 0
        if fname:
            flags = FNAME
        self.fileobj.write(chr(flags))
        write32u(self.fileobj, long(time.time()))
        self.fileobj.write('\002')
        self.fileobj.write('\377')
        if fname:
            self.fileobj.write(fname + '\000')

    def _read_gzip_header(self, f):
        magic = f.read(2)
        if magic != '\037\213':
            raise IOError, 'Not a gzipped file'
        method = ord( f.read(1) )
        if method != 8:
            raise IOError, 'Unknown compression method'
        flag = ord( f.read(1) )
        # modtime = self.fileobj.read(4)
        # extraflag = self.fileobj.read(1)
        # os = self.fileobj.read(1)
        f.read(6)

        if flag & FEXTRA:
            # Read & discard the extra field, if present
            xlen = ord(f.read(1))
            xlen = xlen + 256*ord(f.read(1))
            f.read(xlen)
        if flag & FNAME:
            # Read and discard a null-terminated string containing the filename
            while True:
                s = f.read(1)
                if not s or s=='\000':
                    break
        if flag & FCOMMENT:
            # Read and discard a null-terminated string containing a comment
            while True:
                s = f.read(1)
                if not s or s=='\000':
                    break
        if flag & FHCRC:
            f.read(2)     # Read & discard the 16-bit header CRC


    def write(self,data):
        if self.mode != WRITE:
            import errno
            raise IOError(errno.EBADF, "write() on read-only GzipFile object")

        if self.fileobj is None:
            raise ValueError, "write() on closed GzipFile object"
        if len(data) > 0:
            self.size = self.size + len(data)
            self.crc = zlib.crc32(data, self.crc) & 0xffffffffL
            self.fileobj.write( self.compress.compress(data) )
            self.offset += len(data)

    def _read_eof(self):
        if len(self.decompobj.unused_data) < 8:
            unused = self.decompobj.unused_data  + \
                     self.fileobj.read(8-len(self.decompobj.unused_data))
            if len(unused)<8:
                raise IOError, "Unexpected EOF"
        else:
            unused = self.decompobj.unused_data
        crc32 = struct.unpack(
            "<L", unused[0:4])[0] #le signed int
        isize = struct.unpack(
            "<L", unused[4:8])[0] #le unsigned int

        # Do the CRC check with a LOWU32 on the read self.crcval due to the
        # signed/unsigned problem of the zlib crc code.
        if crc32 != LOWU32(self.crcval):
            raise IOError, "CRC check failed"
        if isize != LOWU32(self.size):
            raise IOError, "Incorrect length of data produced"
        if len(unused) > 8:
            f = PaddedFile(unused[8:], self.fileobj)
        else:
            f = self.fileobj
        try:
            self._read_gzip_header(f)
        except IOError:
            return ''

        if len(unused) > 8:
            return f.unused()
        else:
            return ''
       

    def _read(self, readsize):
        """Bug was here with bad EOF signal"""
        data = self.fileobj.read(readsize)
        is_eof = True
 
        while True:

                
            if data == "":
                decompdata = self.decompobj.flush()
            else:
                decompdata = self.decompobj.decompress(data)
            decomplen = len(decompdata)
            self.buffer.append(decompdata)
            self.bufferlen += decomplen
            self.size += decomplen
            if decomplen:
                is_eof = False
            self.crcval = zlib.crc32(decompdata, self.crcval)
            if self.decompobj.unused_data:
                data = self._read_eof()
                self.decompobj = zlib.decompressobj(-zlib.MAX_WBITS)
                self.crcval = zlib.crc32("")
                self.size = 0
                if data:
                    continue
            break
        return is_eof
 
    def read(self, size=-1):
        """Decompress up to bytes bytes from input.
 
        Raise IOError."""
 
        if self.mode != READ:
            import errno
            raise IOError(errno.EBADF, "read() on write-only GzipFile object")
 
        if self._new_member:
            self._read_gzip_header(self.fileobj)
            self._new_member = False
 
        while size < 0 or self.bufferlen <  size:
            if size < 0:
                readsize = 65536 - self.bufferlen
            else:
                readsize = size - self.bufferlen
                
            if readsize > 65536:
                readsize = 32768
            elif readsize > 32768:
                readsize = 16384
            elif readsize > 16384:
                readsize = 8192
            elif readsize > 8192:
                readsize = 4096
            elif readsize > 4096:
                readsize = 2048
            else:
                readsize = 1024
 
            eof = self._read(readsize)
            if eof:
                break
        if size < 0:
            size = self.bufferlen
        retdata = ""
        while size > 0 and self.buffer:
            decompdata = self.buffer[0]
            decomplen = len(decompdata)
            if size+self.pos <= decomplen:
                tmpdata = decompdata[self.pos:size+self.pos]
                retdata += tmpdata
                self.bufferlen -= size
                self.pos += size
                break
            decomplen -= self.pos
            size -= decomplen
            self.bufferlen -= decomplen
            if self.pos != 0:
                retdata += decompdata[self.pos:]
            else:
                retdata += decompdata
            self.pos = 0
            self.buffer.pop(0)
        self.offset += len(retdata)
        return retdata
 

    def close(self):
        if self.fileobj is None:
            return
        if self.mode == WRITE:
            self.fileobj.write(self.compress.flush())
            write32u(self.fileobj, self.crc)
            # self.size may exceed 2GB, or even 4GB
            write32u(self.fileobj, self.size & 0xffffffffL)
            self.fileobj = None
        elif self.mode == READ:
            self.fileobj = None
        if self.myfileobj:
            self.myfileobj.close()
            self.myfileobj = None

    def __del__(self):
        try:
            if (self.myfileobj is None and
                self.fileobj is None):
                return
        except AttributeError:
            return
        self.close()

    def flush(self,zlib_mode=zlib.Z_SYNC_FLUSH):
        if self.mode == WRITE:
            # Ensure the compressor's buffer is flushed
            self.fileobj.write(self.compress.flush(zlib_mode))
        self.fileobj.flush()

    def fileno(self):
        """Invoke the underlying file object's fileno() method.

        This will raise AttributeError if the underlying file object
        doesn't support fileno().
        """
        return self.fileobj.fileno()

    def isatty(self):
        return False

    def tell(self):
        return self.offset

    def rewind(self):
        '''Return the uncompressed stream file position indicator to the
        beginning of the file'''
        if self.mode != READ:
            raise IOError("Can't rewind in write mode")
        self.fileobj.seek(0)
        self._new_member = True
        self.decompobj = zlib.decompressobj(-zlib.MAX_WBITS)
        self.crcval = zlib.crc32("")
        self.buffer = []   # List of data blocks
        self.bufferlen = 0
        self.pos = 0       # Offset of next data to read from buffer[0]
        self.offset = 0
        self.size = 0

    def seek(self, offset, whence=0):
        if whence:
            if whence == 1:
                offset = self.offset + offset
            else:
                raise ValueError('Seek from end not supported')
        if self.mode == WRITE:
            if offset < self.offset:
                raise IOError('Negative seek in write mode')
            count = offset - self.offset
            for i in range(count // 1024):
                self.write(1024 * '\0')
            self.write((count % 1024) * '\0')
        elif self.mode == READ:
            if offset < self.offset:
                # for negative seek, rewind and do positive seek
                self.rewind()
            count = offset - self.offset
            for i in range(count // 1024):
                self.read(1024)
            self.read(count % 1024)

    def readline(self, size=-1):
        if self._new_member:
            self._read_gzip_header(self.fileobj)
            self._new_member = False
 
        scansize = 0        
        buffpos = 0
        
        while True:
            for idx in range(buffpos, len(self.buffer)):
                if idx == 0:
                    scansize -= self.pos
                    pos = self.buffer[idx].find('\n', self.pos)
                else:
                    pos = self.buffer[idx].find('\n')                    
                if pos != -1:
                    if size>=0 and scansize+pos+1>size:
                        return self.read(size)
                    return self.read(scansize+pos+1)
                scansize += len(self.buffer[idx])
                if size>=0 and scansize>size:
                    return self.read(size)
            buffpos = len(self.buffer)
            eof = self._read(1024)
            if eof:
                return self.read(scansize)

    def readlines(self, sizehint=0):
        # Negative numbers result in reading all the lines
        if sizehint <= 0:
            sizehint = sys.maxint
        L = []
        while sizehint > 0:
            line = self.readline()
            if line == "":
                break
            L.append(line)
            sizehint = sizehint - len(line)

        return L

    def writelines(self, L):
        for line in L:
            self.write(line)

    def __iter__(self):
        return self

    def next(self):
        line = self.readline()
        if line:
            return line
        else:
            raise StopIteration


def _test():
    # Act like gzip; with -d, act like gunzip.
    # The input file is not deleted, however, nor are any other gzip
    # options or features supported.
    args = sys.argv[1:]
    decompress = args and args[0] == "-d"
    if decompress:
        args = args[1:]
    if not args:
        args = ["-"]
    for arg in args:
        if decompress:
            if arg == "-":
                f = GzipFile(filename="", mode="rb", fileobj=sys.stdin)
                g = sys.stdout
            else:
                if arg[-3:] != ".gz":
                    print "filename doesn't end in .gz:", repr(arg)
                    continue
                f = open(arg, "rb")
                g = __builtin__.open(arg[:-3], "wb")
        else:
            if arg == "-":
                f = sys.stdin
                g = GzipFile(filename="", mode="wb", fileobj=sys.stdout)
            else:
                f = __builtin__.open(arg, "rb")
                g = open(arg + ".gz", "wb")
        while True:
            chunk = f.read(1024)
            if not chunk:
                break
            g.write(chunk)
        if g is not sys.stdout:
            g.close()
        if f is not sys.stdin:
            f.close()

if __name__ == '__main__':
    _test()



"""
Top-level object for PiCloud
Manages caching and client-side type checking
"""
from __future__ import with_statement
"""
Copyright (c) 2011 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@piscicloud.com

The scicloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see 
http://www.gnu.org/licenses/lgpl-2.1.html
"""
 
import sys
import threading
import time
import cPickle as pickle
import collections
import weakref
import types

"""Debug"""
DEBUG = False
if DEBUG:
    import signal
    import traceback
    import os

    def dumpstacks():
        id2name = dict([(th.ident, th.name) for th in threading.enumerate()])
        code = []
        for threadId, stack in sys._current_frames().items():
            code.append("\n# Thread: %s(%d)" % (id2name.get(threadId), threadId))
            for filename, lineno, name, line in traceback.extract_stack(stack):
                code.append('File: "%s", line %d, in %s' % (filename, lineno, name))
                if line:
                    code.append("  %s" % (line.strip()))
        return '\n'.join(code)
    
    def savestacks(sig, frame):
        print 'debugging..'
        contents = dumpstacks()
        f = open('/tmp/scicloud_%s' % os.getpid(), 'wb')
        f.write(contents)
        f.close()    
    
    signal.signal(signal.SIGUSR1, savestacks)
    print 'registered debug for %s' % os.getpid()


try:
    import json    
except ImportError: #If python version < 2.6, we need to use simplejson
    import simplejson as json

try:
    from itertools import izip_longest
except ImportError: #python 2.5 lacks izip_longest
    from .util import izip_longest 
from itertools import izip, chain

from . import scicloudconfig as cc
from .scicloudlog import scicloudLog
from . import serialization

from .util import funcname, validate_func_arguments, fix_time_element 
from .util.cache import JobCacheManager, JobFiniteDoubleCacheManager, \
    JobFiniteSizeCacheManager, JobAbstractCacheManager
from .util.xrange_helper import filter_xrange_list, maybe_xrange_iter


# set by the instantiation of a Cloud object
_scicloud = None

class CloudException(Exception):
    """
    Represents an exception that has occurred by a job on the scicloud.
    CloudException will display the job id for failed scicloud requests.
    """
    
    status = None
    jid = None
    parameter = None
    hint = None
    
    job_str = 'Job '
    status_str = 'Error '
    
    def __init__(self, value, jid=None, status=None, retry=False, logger=scicloudLog, hint=None):
        self.jid = jid
        self.parameter = value
        self.status = status
        self.retry = retry
        if logger:
            self._logmsg(logger)
        self.hint = hint
        super(CloudException, self).__init__(jid, status, value, jid)
        
    def __str__(self):
        if self.jid is not None:
            res = self.job_str + str(self.jid) + ': ' + str(self.parameter)
            if self.hint:
                res += '\nTIP: %s' % self.hint
            return res
        elif self.status is not None:
            return self.status_str + str(self.status) + ': ' + str(self.parameter)
        else:
            return str(self.parameter)      

    def _logmsg(self, logger):
        """logmsg - only log on jid related"""
        if self.jid is not None:
            logger.warning(''.join(['Job ',str(self.jid),' threw exception:\n ', str(self.parameter)]))
            
class CloudTimeoutError(CloudException):
    """CloudException specifically for join timeouts"""
    
    def __init__(self, value = 'Timed out', jid=None, logger = scicloudLog):
        CloudException.__init__(self, value, jid, None, logger)
    
    
    def _logmsg(self, logger):
        """logmsg - only log on jid related"""
        if self.jid is not None:
            logger.warning(''.join(['Job ',str(self.jid),' %s\n ' % self.parameter ]))

def _getExceptionMsg(status):
    """Return exception method associated with a status"""
    if status == 'killed':
        return 'Killed'
    elif status == 'stalled':
        return 'Could not start due to dependency erroring'
    else:
        return 'unknown error'      

class FakeException(Exception):
    pass

#_catchCloudException = CloudException
_catchCloudException = FakeException
_docmsgpkl = "\nPlease see PiCloud documentation for more information about pickle errors"

class Cloud(object):
    """
    Top-level object that manages client-side type checking and client-side cache for Cloud  
    """
       
    job_cache_size = 8191 #cache size for cache manager. None = no cache; 0 = no limit
    result_cache_size = 4096 #cache size for results (applies only if limited jobCache) in bytes; None = no cache; 0 = no limit
    
    finished_statuses = ['done', 'stalled', 'error', 'killed']
    
    # process locally or over network via http
    adapter = None
    
    # manages callbacks
    manager = None
    
    #manages job caching
    cacheManager = None

    __isopen = False
    
    openLock = None
    
    parentModule = None #string representation of module this object is placed in (if any)
    
    __running_on_scicloud = cc.transport_configurable('running_on_scicloud', default=False, 
                                        comment="Internal. States if running on PiCloud",
                                        hidden = True) 
    
    """compute type constants"""
    c1 = 'c1'
    c2 = 'c2'
    f2 = 'f2'
    m1 = 'm1'
    s1 = 's1'
        
    @property
    def opened(self): 
        """Returns whether the scicloud is open"""
        return self.__isopen
    
    @classmethod
    def running_on_scicloud(cls):  #high level check
        """
        Returns true iff current thread of execution is running on PiCloud servers.
        """
        
        return cls.__running_on_scicloud
           
    def __init__(self, adapter):
        self.adapter = adapter
        adapter._scicloud = self
        self.openLock = threading.RLock()
        self._opening = False
    
    def open (self):
        """Enable the scicloud"""
        
        with self.openLock:
            
            if self.opened: #accept this for threading issues
                return                
            
            #hackish re-entry check -- due to issues with local adapter opening this           
            if self._opening:
                return

            try:
                self._opening = True
                if not self.adapter.opened:
                    self.adapter.open()                    
            except _catchCloudException, e:
                raise e
            finally:
                self._opening = False
    
            self.manager = CloudTicketManager(self)
            self.manager.start() 
            
            if self.job_cache_size == None: #no cache
                self.cacheManager = JobAbstractCacheManager()
            elif self.job_cache_size == 0: #no limit on cache
                self.cacheManager = JobCacheManager()            
            else:
                if self.result_cache_size == None:  #no cache
                    resultCacheSize = 0
                elif self.result_cache_size == 0:  #effectively no limit on cache
                    resultCacheSize = sys.maxint
                else:                           #default
                    resultCacheSize = self.result_cache_size
                
                childManager = JobFiniteSizeCacheManager(resultCacheSize,'result')             
                self.cacheManager = JobFiniteDoubleCacheManager(self.job_cache_size, childManager)
                 
            self.__isopen = True
            if not self.adapter.isSlave:
                scicloudLog.info('Cloud started with adapter =%s', str(self.adapter))  
    
    
    def close(self):
        """
        Terminate scicloud connection (or simulation) and all threads utilized by scicloud.
        Returns True if closed scicloud; False if scicloud was already closed 
        
        .. note::
        
            scicloud will automatically re-open if you invoke any scicloud.* again
            
        .. warning::
        
            This function should never be called by a job running on PiCloud 
        """    
        if not self.opened:
            return False                   
        if self.adapter.opened:
            self.adapter.close()    
        self.manager.stop()
        scicloudLog.info('Cloud closed with adapter =%s', str(self.adapter))
        self.__isopen = False
        return True                
    
    def _checkOpen(self):
        """Open scicloud if it is not already"""
        if not self.opened:
            self.open()
    
    def is_simulated(self):
        """
        Returns true iff scicloud processor is being run on the local machine.
        """        
        return self.adapter.is_simulated() 
    
    def connection_info(self):
        """
        Returns a dictionary of information describing the current connection.
        """
        return self.adapter.connection_info()

    
    def needs_restart(self, **kwargs):        
        """
        For internal use
        Return true if scicloud requires restart based on changed params
        """
        
        if not self.__isopen: #by definition must restart
            return True
        
        return self.adapter.needs_restart(**kwargs)
    
    def _getJobParameters(self, func, mappings, ignore=[]):
        """Return parameters from function and mappings
        Removes keys from mappings handled by this function
        Ignores keys in ignore[]"""
        
        # defaults
        job_label = None
        job_priority = 5
        job_restartable = True
        _type = 'c1'
        cores = 1
        profile = False
        depends_on = []
        depends_on_errors = 'abort'
        fast_serialization = 0
        kill_process = False #not documented yet
        env = None
        vol = []
        max_runtime = None
        os_env_vars = []
        
        if '_label' in mappings and '_label' not in ignore:
            #job_label must be an ASCII string
            try:
                job_label = mappings['_label'].decode('ascii').encode('ascii')
            except (UnicodeDecodeError, UnicodeEncodeError):
                raise TypeError('_job_label must be an ASCII string')
            del mappings['_label']
        
        if '_priority' in mappings and '_priority' not in ignore:
            job_priority = mappings['_priority']
            if not isinstance(job_priority, (int, long)):
                raise TypeError(' _priority must be an integer')             
            del mappings['_priority']

        if '_restartable' in mappings and '_restartable' not in ignore:
            job_restartable = mappings['_restartable']
            if not isinstance(job_restartable, bool):
                raise TypeError(' _job_restartable must be boolean')             
            del mappings['_restartable']
        
        if '_depends_on' in mappings and '_depends_on' not in ignore:
            if hasattr(mappings['_depends_on'],'__iter__'):
                depends_on = mappings['_depends_on']
            elif isinstance (mappings['_depends_on'], (int,long)):
                depends_on = [mappings['_depends_on']]
            else:
                raise TypeError('_depends_on must be a jid or list of jids')
            del mappings['_depends_on']
            
            # Validate depends_on
            if not isinstance(depends_on, xrange):
                test = reduce(lambda x, y: x and (isinstance(y,(int,long,xrange))), depends_on, True)
                if not test:
                    raise TypeError( '_depends_on list can only contain jids' )

        if '_depends_on_errors' in mappings and '_depends_on_errors' not in ignore:
            if isinstance(mappings['_depends_on_errors'], basestring):
                depends_on_errors = str(mappings['_depends_on_errors'])
            else:
                raise TypeError('_depends_on_errors must be a string')
            if depends_on_errors not in ['ignore', 'abort']:
                raise ValueError('_depends_on_errors must be "abort" or "ignore"')
            del mappings['_depends_on_errors']
        
        if '_fast_serialization' in mappings  and '_fast_serialization' not in ignore:
            fast_serialization = mappings['_fast_serialization']
            
            if not isinstance(fast_serialization, (int, long)):
                raise TypeError(' _fast_serialization must be an integer')
            elif isinstance(fast_serialization, types.BooleanType):
                raise TypeError(' _fast_serialization must be an integer')
                                    
            del mappings['_fast_serialization']    
        
        if '_type' in mappings and '_type' not in ignore:
            if isinstance(mappings['_type'], basestring):
                _type = str(mappings['_type'])            
            else:
                raise TypeError('_type must be a string')
            del mappings['_type']
        
        if '_high_cpu' in mappings  and '_high_cpu' not in ignore:
            #TODO: Warn that this is deprecated
            if isinstance(mappings['_high_cpu'], bool):
                high_cpu = mappings['_high_cpu']
                if high_cpu:
                    _type = 'c2'
            else:
                raise TypeError('_high_cpu must be a boolean')
            del mappings['_high_cpu']
        
        if '_cores' in mappings  and '_cores' not in ignore:
            if isinstance(mappings['_cores'], (int, long)) and mappings['_cores'] > 0:
                cores = mappings['_cores']
            else:
                raise TypeError('_cores must be a positive integer')
            del mappings['_cores']
            
        if '_env' in mappings and  '_env' not in ignore:
            env = str(mappings['_env'])
            del mappings['_env']
            
        if '_vol' in mappings and  '_vol' not in ignore:
            vol = mappings['_vol']
            if not hasattr(vol, '__iter__'):
                vol = [vol]
            for e in vol:
                if not isinstance(e, basestring):
                    raise TypeError('_vol must be a string or a list of strings')
            del mappings['_vol']

        if '_max_runtime' in mappings and  '_max_runtime' not in ignore:
            if isinstance( mappings['_max_runtime'], (int, long) ):
                max_runtime = mappings['_max_runtime']
                if max_runtime <= 0:
                    raise ValueError('_max_runtime must be an integer > 0')
            else:
                raise TypeError('_max_runtime must be an integer')
            
            del mappings['_max_runtime']

        if '_profile' in mappings and  '_profile' not in ignore:
            profile = bool( mappings['_profile'])
            del mappings['_profile']
            
        if '_kill_process' in mappings and '_kill_process' not in ignore:
            kill_process = bool( mappings['_kill_process'])
            del mappings['_kill_process']

        if '_os_env_vars' in mappings and '_os_env_vars' not in ignore:
            os_env_vars =  mappings['_os_env_vars']
            if not hasattr(os_env_vars, '__iter__'):
                raise TypeError('_os_env_vars must be an iterable')
            del mappings['_os_env_vars']
        
        parameters = {'fast_serialization': fast_serialization,
                      'func_name': funcname(func) if func else None,
                      'label': job_label,
                      'priority': job_priority,
                      'restartable': job_restartable,
                      'type': _type,
                      'cores': cores,
                      'env' : env,
                      'vol' : vol,
                      'max_runtime': max_runtime,
                      'profile': profile,
                      'depends_on': depends_on,
                      'depends_on_errors' : depends_on_errors,
                      'kill_process' : kill_process,
                      'os_env_vars' : os_env_vars
                      }
        
        for ig_key in ignore:
            if ig_key in parameters:
                del parameters[ig_key]
        
        return parameters
    
    def _check_jid_type(self, jids):
        if not isinstance(jids, xrange):
            test = reduce(lambda x, y: x and (isinstance(y,(int,long,xrange))), jids, True)
            if not test:
                raise TypeError( 'jid(s) must be integers' )
    
    def call(self, func, *args, **kwargs):
        """
        Create a 'job' that will invoke *func* (a callable) in the scicloud.
        
        When the job is run, *func* will be invoked on PiCloud with the
        specified *args* and all non-reserved *kwargs*.

        Call will return an integer Job IDentifier (jid) which can be passed into 
        the status and result methods to obtain the status of the job and the 
        job result.
        
        Example::    
        
            def add(x, y):
                return x + y
            scicloud.call(add, 1, 2) 
        
        This will create a job that when processed will invoke `add` with x=1 and y=2.
        
        See online documentation for additional information.
        
        Reserved special *kwargs* (see docs for details):
            
        * _callback: 
            A list of functions that should be run on the callee's computer once
            this job finishes successfully.
            Each callback is passed one argument: the jid of the complete job
        * _callback_on_error:  
            A list of functions that should be run on the callee's computer if this
            job errors.  
        * _cores:
            Set number of cores your job will utilize. See http://docs.scivm.com/primer.html#choose-a-core-type/
            In addition to having access to more CPU cores, the amount of RAM available will grow linearly.
            Possible values for ``_cores`` depend on what ``_type`` you choose:
            
            * c1: 1                    
            * c2: 1, 2, 4, 8
            * f2: 1, 2, 4, 8, 16                                    
            * m1: 1, 2, 4, 8
            * s1: 1        

        * _depends_on:
            An iterable of jids that represents all jobs that must complete successfully 
            before the job created by this call function may be run.
        * _depends_on_errors:
            A string specifying how an error with a jid listed in _depends_on should be handled:       
                 
            * 'abort': Set this job to 'stalled' (Default)
            * 'ignore': Treat an error as satisfying the dependency
        * _env:
            A string specifying a custom environment you wish to run your job within.
            See environments overview at http://docs.scivm.com/environment.html                
        * _fast_serialization:
            This keyword can be used to speed up serialization, at the cost of some functionality.
            This affects the serialization of both the arguments and return values            
            Possible values keyword are:
                        
            0. default -- use scicloud module's enhanced serialization and debugging info            
            1. no debug -- Disable all debugging features for arguments            
            2. use cPickle -- Use Python's fast serializer, possibly causing PicklingErrors
            
            ``func`` will always be serialized by the enhanced serializer (with debugging info).                
                    
        * _kill_process:
            Terminate the jobs' Python interpreter after *func* completes, preventing
            the interpreter from being used by subsequent jobs.  See Technical Overview for more info.                
        * _label: 
            A user-defined string label that is attached to the jobs. Labels can be
            used to filter when viewing jobs interactively (i.e. on the PiCloud website).
        * _max_runtime:
            Specify the maximum amount of time (in integer minutes) jobs can run. If the job runs beyond 
            this time, it will be killed.           
        * _os_env_vars:
            List of operating system environment variables that should be copied to PiCloud from your system
            Alternatively a dictionary mapping the environment variables to the desired values.
        * _priority: 
            A positive integer denoting the jobs' priority. PiCloud tries to run jobs 
            with lower priority numbers before jobs with higher priority numbers.                
    	* _profile:
    	    Set this to True to enable profiling of your code. Profiling information is 
    	    valuable for debugging, and finding bottlenecks.
    	    **Warning**: Profiling can slow your job significantly; 10x slowdowns are known to occur
        * _restartable:
            In the rare event of hardware failure while a job is processing, this flag
            indicates whether the job can be restarted. By default, this is True. Jobs
            that modify external state (e.g. adding rows to a database) should set this
            False to avoid potentially duplicate modifications to external state.
        * _type:
            Choose the type of core to use, specified as a string:
            
            * c1: 1 compute unit, 300 MB ram, low I/O (default)                    
            * c2: 2.5 compute units, 800 MB ram, medium I/O
            * f2: 5.5 compute units, 3.75 GB ram, high I/O, hyperthreaded core                                    
            * m1: 3.25 compute units, 8 GB ram, high I/O
            * s1: Up to 2 compute units (variable), 300 MB ram, low I/O, 1 IP per core
                           
            See http://www.scivm.com/pricing/ for pricing information
        * _vol:
            A string or list of strings specifying a volume(s) you wish your jobs to have access to.
        """            
        
        self._checkOpen()

        if not callable(func):
            raise TypeError( 'scicloud.call first argument (%s) is not callable'  % (str(func) ))
        
        parameters = self._getJobParameters(func, kwargs)
        
        if '_callback' in kwargs:
            callback = kwargs['_callback']
            del kwargs['_callback']
        else:
            callback = None
            
        if '_callback_on_error' in kwargs:
            callback_on_error = kwargs['_callback_on_error']        
            del kwargs['_callback_on_error']
        else:
            callback_on_error = None       
            
        validate_func_arguments(func, args, kwargs) 
        
        try:
            jid = self.adapter.job_call(parameters,func,args,kwargs)
        except _catchCloudException, e:
            raise e    
        except pickle.PicklingError, e:            
            e.args = (e.args[0] + _docmsgpkl,)            
            raise e
              
        # only add ticket to manager (will poll to see whether job has
        # finished) if there is a callback assigned to the call
        if callback:            
            self.manager.add_callbacks(jid, callback, 'success')
        if callback_on_error:
            self.manager.add_callbacks(jid, callback_on_error, 'error')        

        return jid    
    
    def join(self, jids, timeout = None, ignore_errors = False, deadlock_check=True):
        """
        Block current thread of execution until the job specified by the integer *jids*
        completes.  Completion is defined as the job finishing or erroring
        (including stalling).          
        If the job errored, a CloudException detailing the exception that triggered
        the error is thrown. If the job does not exist, a CloudException is thrown.
        
        This method also accepts an iterable describing *jids* and blocks until all
        corresponding jobs finish. If an error is seen, join may terminate before 
        all jobs finish.  If multiple errors occur, it is undefined which job's exception
        will be raised.
        
        If *timeout* is set to a number, join will raise a CloudTimeoutError if the
        job is still running after *timeout* seconds
        
        If *ignore_errors* is True, no CloudException will be thrown if a job errored.
        Join will block until every job is complete, regardless of error status.
        
        If *deadlock_check* is True (default), join will error if it believes deadlock 
        may have occured, as described in our  
        `docs <http://docs.scivm.com/scicloud_pitfalls.html#scicloud-call-recursion>`_
        """
        
        #TODO: Use scicloud ticket manager with this
        poll_interval = 1.0
        
        """
        On PiCloud (and in simulation), a job waiting on another job can cause live lock 
        (job being waited on cannot start due to lack of available resources)
        We detect this by requiring that the jobs we are waiting on switch from queue to processing w/in 60 s
        If this fails to happen, we consider a deadlock to have occured"""
        max_recursive_wait = 60.0 # number of seconds to wait for a job to go from queue to processing
        
        self._checkOpen()
                
        if not hasattr(jids,'__iter__'):
            jids = [jids] 
        
        self._check_jid_type(jids)        
        
        #variables filter can modify:
        #NOTE: With Python3, they can just be nonlocal
        #Some of this code is legacy
        class FilterSideEffects:
            aptr = 0                #Pointer into returned statuses/exceptions
            num_queued = 0          #number of jobs with status 'queued'            
        
        def filterJob(jid, status):
            """Helper function to filter out done jobs and fire exceptions if needed                 
            """                        
            if status != 'done': 
                if status in self.finished_statuses:
                    if ignore_errors:
                        return False
                    info_dct = self.info(jid,['exception', '_hint'])[jid]
                    exception = info_dct['exception']
                    hint = info_dct.get('_hint','')
                    if status == 'error' or (status == 'killed' and exception):
                        raise CloudException(exception, jid=jid, status=status, hint=hint)
                    else:
                        msg = _getExceptionMsg(status)
                        raise CloudException(msg, jid=jid, status=status)  
                return True  #keep iterating
            return False            
       
        cachedStatuses = self.cacheManager.getCached(jids, ('status',))
       
        def cacheJobFilter(jid):
            """Filter out cached jobs"""
            status = cachedStatuses.get(jid,{}).get('status',None)
            if status:
                return filterJob(jid, status)
            else: #job not in cache
                return True  
        
        neededJids = filter_xrange_list(cacheJobFilter,jids)

        start_time = time.time()
        can_recurse_timeout = False
        latest_job_started_time = time.time() # set to time.time() on queue-->processing transition
        old_queued = 0 # number of jobs seen as queued on last iteration
        adapter_has_join = True # does the adapter support jobs_join?
        
        while True:
            
            # due to lag with status testing, only timeout if in timeout condition before test
            can_timeout = timeout and (time.time() > start_time + timeout) 
            if deadlock_check and (self.running_on_scicloud() or self.adapter.isSlave):
                can_recurse_timeout = (time.time() > latest_job_started_time + max_recursive_wait)
            
            if neededJids:
                try:
                    effective_timeout = 0 if (can_timeout or can_recurse_timeout) else min(60,timeout)
                    astatuses = self.adapter.jobs_join(neededJids, effective_timeout)
                    if astatuses is False: #adapter does not support jobs_join:
                        adapter_has_join = False
                        astatuses = self.status(neededJids)
                except _catchCloudException, e:
                    raise e                             
            else:
                break 
            
            fse = FilterSideEffects() #used by below function                        
            def resultFilter(jid):
                """Filter out jobs we got the status for"""            
                status = astatuses[fse.aptr]
                fse.aptr += 1                
                if status == 'queued':
                    fse.num_queued+=1
                return filterJob(jid, status) # filter out done jobs; raise exception on errored jobs                                
            
            neededJids = filter_xrange_list(resultFilter, neededJids)            
            
            if not neededJids:
                break

            if can_timeout:
                raise CloudTimeoutError('scicloud.join timed out', neededJids)
            
            # live lock handling
            if old_queued != fse.num_queued:
                latest_job_started_time = time.time()
                old_queued = fse.num_queued
            elif can_recurse_timeout and old_queued:
                raise CloudTimeoutError('scicloud.join detected possible dead-lock due to recursive jobs. See http://docs.scivm.com/scicloud_pitfalls.html#scicloud-call-recursion', neededJids)
            
            if not adapter_has_join:
                time.sleep(poll_interval) #poll
            
    def status(self, jids):
        """
        Returns the status of the job specified by the integer *jids*. If 
        the job does not exist, a CloudException is thrown.
        This method also accepts an iterable describing *jids*, in which case a respective list
        of statuses is returned."""
        #really a wrapper for scicloud.info
        deseq = False
        if not hasattr(jids,'__iter__'):
            jids = [jids]
            deseq = True
                
        status_dict = self.info(jids,'status')
        
        if deseq:
            return status_dict[jids[0]]['status']
        else:
            return [status_dict[jid]['status'] for jid in maybe_xrange_iter(jids)]
        
    def result(self, jids, timeout = None, ignore_errors = False, deadlock_check=True):
        """
        Blocks until the job specified by the integer *jids* has completed and
        then returns the return value of the job.         
        If the job errored, a CloudException detailing the exception that triggered
        the error is thrown. 
        If the job does not exist, a CloudException is thrown.        

        This function also accepts an iterable describing *jids*, in which case a respective list
        of return values is returned.
        
        If *timeout* is set to a number, result will raise a CloudTimeoutError if the
        job is still running after *timeout* seconds
        
        If *ignore_errors* is True, a job that errored will not raise an exception.
        Instead, its return value will be the CloudException describing the error.
        
        If *deadlock_check* is True (default), result will error if it believes deadlock 
        may have occured, as described in our  
        `docs <http://docs.scivm.com/scicloud_pitfalls.html#scicloud-call-recursion>`_
        """        
        return self.__result(jids, timeout, ignore_errors, deadlock_check)
 
    def iresult(self, jids, timeout = None, num_in_parallel=10, ignore_errors = False):
        """
        Similar to result, but returns an iterator that iterates, in order, through  
        the return values of *jids* as the respective job finishes, allowing quicker 
        access to results and reducing memory requirements of result reading.
        
        If a job being iterated over errors, an exception is raised. 
        However, the iterator does not exhaust; if the exception is caught, one can continue
        iterating over the remaining jobs.
        
        If *timeout* is set to a number, a call to the iterator's next function 
        will raise a CloudTimeoutError after *timeout* seconds if no result becomes available.
        
        *num_in_parallel* controls how many results are read-ahead from the scicloud
        Set this to 0 to use the allowed maximum.
        
        If *ignore_errors* is True, a job that errored will return the CloudException describing 
        its error, rather than raising an Exception        
        """        
        return self.__iresult(jids, timeout, num_in_parallel, ignore_errors)
         
    def __result(self, jids, timeout=None, ignore_errors=False, deadlock_check=True, by_jid=False):
        deseq = False
        if not hasattr(jids,'__iter__'):
            jids = [jids]
            deseq = True

        self._check_jid_type(jids)

        self._checkOpen()

        if not by_jid:
            try:
                # wait for jids to be ready
                self.join(jids, timeout = timeout, 
                          ignore_errors=ignore_errors, deadlock_check=deadlock_check)  
            except _catchCloudException, e:
                raise e            

        #everything is now cached AND done
        if by_jid:
            cached_results = {}
        else:
            cached_results  = self.cacheManager.getCached(jids, ('result',))        
        needed_jids = filter_xrange_list(lambda jid: jid not in cached_results,jids)
        
        jid_exceptions = {}
        
        if ignore_errors:
            #as ignore_errors is a rarely used flag, this block is not fully optimized                        
            exceptions_dct = self.info(jids, ['exception', 'status'])
            
            for jid, excp_dct in exceptions_dct.iteritems():
                status = excp_dct['status']
                if status != 'done':
                    exception = excp_dct['exception']
                    if not exception:
                        exception = _getExceptionMsg(status)
                    jid_exceptions[jid] = (status, exception)
            needed_jids = filter_xrange_list(lambda jid: jid not in jid_exceptions, jids)                 

        if needed_jids:
            try:
                resp = self.adapter.jobs_result(jids=needed_jids, by_jid=by_jid)
                
            except _catchCloudException, e:
                raise e                                           
            
            results_list = resp['data']
            interpretation = resp['interpretation']

            aresults = izip_longest(interpretation, results_list)

        else:
            aresults = []

        a_result_iter = aresults.__iter__()
        
        outresults = []
        
        for jid in maybe_xrange_iter(jids):
            
            if cached_results.has_key(jid):            
                result = cached_results.get(jid)['result']
                outresults.append(serialization.deserialize(result))
            
            elif jid_exceptions.has_key(jid):
                status, exception = jid_exceptions.get(jid)
                clexp = CloudException(exception, jid=jid, status=status, logger=None)                
                outresults.append(clexp)
                continue
            
            else:
                interpret, result = a_result_iter.next()

                if interpret['datatype'] == 'python_pickle':
                    outresults.append(serialization.deserialize(result))
                    if not by_jid:
                        self.cacheManager.putCached(jid, status='done', result=result)
                                                
                elif interpret['datatype'] == 'json':
                    if interpret['action'] == 'scicloud.iresult_global':
                        sub_jids = json.loads(result)
                        sub_results_iter = chain.from_iterable( self.__iresult(sub_jids, num_in_parallel=1, by_jid=True) )
                        outresults.append(sub_results_iter)
                        
        if deseq:
            return outresults[0]
        else:
            return outresults 
   
    def __iresult(self, jids, timeout = None, num_in_parallel=10, ignore_errors = False, by_jid=False):
        if not hasattr(jids,'__iter__'):
            jids = [jids]
        
        num_jids = len(jids)
        
        max_in_parallel = 1024
        if num_in_parallel < 1 or num_in_parallel > max_in_parallel:
            num_in_parallel = max_in_parallel
        
        self._checkOpen()
                            
        ready_results = collections.deque()  #queue of results
        result_cv = threading.Condition()  #both main and loop threads sleep on this
        errorJid = []  #1st element set to a job if an error is to be returned
        isDone = [] #1st element set to True when result_loop finishes. Set to an exception if result_loop crashed
        
        def result_loop(iterator_ref):
            """
            Helper function run inside another thread
            It downloads results as the main thread continues work
            
            Slight issue is present in that if the condition variable is not necessarily
            waited on if an error is present in the ready_results buffer
            """
            jidIterator = maybe_xrange_iter(jids)
            jids_testing = collections.deque()   #queue of jobs being tested            
            poll_interval = 1.0
            
            try:
                while self.__isopen:                
                    #First add jobs to check
                    try:
                        while len(jids_testing) < num_in_parallel:
                            jids_testing.append(jidIterator.next())
                    except StopIteration:
                        if not jids_testing:  #loop is done!
                            scicloudLog.debug('scicloud.iresult.result_loop has completed')
                            isDone.append(True)
                            break 
                    
                    with result_cv:
                        while True:
                            if not iterator_ref(): #GC Detection on iterator removal
                                scicloudLog.debug('scicloud.iresult.result_loop detected garbage collection')
                                break
                            
                            if (len(ready_results) < num_in_parallel and not errorJid):
                                break
                            result_cv.wait(5.0)  #poll for Garbage collection
                      
                    if not iterator_ref():  #GC Detection on iterator removal
                        break
                    assert (len(ready_results) < num_in_parallel)
                    assert (not errorJid)
                         
                    if by_jid:
                        statuses = ['done' for _ in xrange(num_jids)]
                    else:
                        statuses = self.status( jids_testing )
                        scicloudLog.debug('scicloud.__iresult.result_loop testing %s. got status %s' % \
                                       (jids_testing, statuses))
                    ctr = 0
                    jids_to_access = []
                
                    #find all done jobs whose results we can grab
                    for jid in jids_testing:                    
                        status = statuses[ctr]
                        if status not in self.finished_statuses:
                            break
                        if status == 'done':                    
                            jids_to_access.append(jid)
                        else: 
                            #postpone error handling until all good results handled:
                            if not jids_to_access and not ready_results:                                                                                 
                                errorJid.append(jid)
                            break
                        ctr+=1
                    #technically race condition here -- what if error_jid is popped?
                    if errorJid: #on error, return to top. Top will wait on cv
                        jids_testing.popleft()
                        with result_cv:
                            result_cv.notify()
                        continue 
                    
                    if not jids_to_access: #no jobs ready yet -- wait and return to top
                        time.sleep(poll_interval)  
                        #FIXME: Should use adapter.join
                        continue                
                
                    if not by_jid:
                        scicloudLog.debug('scicloud.__iresult.result_loop getting results of %s' % jids_to_access)

                    new_results  = self.__result(jids_to_access, by_jid=by_jid) 
                    with result_cv:
                        for result in new_results:
                            ready_results.append(result)
                            jids_testing.popleft()
                        result_cv.notify()

                scicloudLog.debug('scicloud.__iresult.result_loop is terminating')
                with result_cv:
                    result_cv.notify()
            
            except Exception, e:
                scicloudLog.exception('scicloud.iresult.result_loop crashed')
                isDone.append(e)
                with result_cv:
                    result_cv.notify()
 

        class ResultIterator(object):
            """Iterator that handles results"""
            
            def __init__(self, parentInstance):
                self.parent = parentInstance              
                self.cnt = 0
                self.crashed = False
            
            def __iter__(self):
                return self
            
            def _checkDone(self):
                if isDone:
                    scicloudLog.debug('scicloud.__iresult.ResultIterator is done.')
                    if isinstance(isDone[0], Exception):  #something went wrong
                        scicloudLog.error('ResultIterator detected that polling thread crashed')
                        self.crashed = True #if next is ever called again, raise StopIteration
                        raise isDone[0]
                    raise StopIteration
                
            def next(self, timeout = timeout):
                """Just walk through ready_results"""     
                
                if self.crashed:
                    raise StopIteration
          
                with result_cv:                    
                    if not errorJid and not ready_results:
                        self._checkDone()
                        scicloudLog.debug('scicloud.__iresult.ResultIterator is going to sleep')
                        result_cv.wait(timeout)
                        if not errorJid and not ready_results:
                            self._checkDone()
                            
                            #we timed out
                            raise CloudTimeoutError('__iresult iterator timed out')
                        scicloudLog.debug('scicloud.__iresult.ResultIterator is awake. num ready results = %s errorJid is %s' % (len(ready_results), errorJid))
                    
                    result_cv.notify() #waken result_loop thread
                    
                    if errorJid:  #trigger exception with join:
                        try:
                            self.parent.join(errorJid)
                        except CloudException, e:
                            errorJid.pop()  #result thread can now continue
                            if ignore_errors:
                                return e
                            else:
                                raise e
                        else: #should never occur
                            assert(False) 
                        
                    assert(ready_results)
                    return ready_results.popleft()
                    
        ri = ResultIterator(self)
        result_thread = threading.Thread(target=result_loop, args=(weakref.ref(ri),))  
        result_thread.daemon = True
        result_thread.name = 'scicloud.iresult.result_loop'
        result_thread.start()  
        
        return ri
    
    def iresult_unordered(self, jids):
        """Hackish - probably should not use"""
        
        class UnorderedResultIterator(object):
            """Iterator that handles results"""
            
            def __init__(self, parent):              
                self.jids_testing = list(maybe_xrange_iter(jids))
                self.parent = parent
                
            
            def __iter__(self):
                return self
            
            def next(self):
                                
                while True:
                    
                    jids = self.jids_testing                
                    statuses = self.parent.status( jids )                
                                
                    for jid, status in zip(jids, statuses):
                        if status in self.parent.finished_statuses:
                            self.jids_testing.remove(jid)
                            
                            if status == 'done': #ignore errors
                                res = self.parent.__result(jid)                            
                                return jid, res
                    
                    if self.jids_testing:
                        time.sleep(0.5)
                    else:
                        raise StopIteration
        
        return UnorderedResultIterator(self)
               
    def info(self, jids, info_requested=None):
        """        
        Request information about jobs specified by integer or iterable *jids*
        
        As this function is designed for console debugging, the return value 
        is a dictionary whose keys are the *jids* requested.  The values are themselves
        dictionaries, whose keys in turn consist of valid values within the iterable
        *info_requested*.  Each key maps to key-specific information
        about the job, e.g. stdout maps to standard output of job.

        Possible *info_requested* items are one of more of:
        
        * status: Job's status
        * stdout: Standard Outout produced by job (last 64k characters)
        * stderr: Standard Error produced by job (last 64k characters)
        * logging: Logs produced with the Python logging module.
        * pilog: Logs generated by PiCloud regarding your job.
        * exception: Exception raised (on error)
        * runtime: runtime (wall clock) of job, specifically how long function took to evaluate
        * created: datetime when job was created on scivm by a call/map
        * finished: datetime when job completed running. note that finished - created = runtime + time in queue
        * env: Environment job runs under
        * vol: list of volumes job runs under
        * function: Name of function
        * label: Label of function        
        * args: Arguments to function
        * kwargs: Keywork arguments to function
        * func_body: Actual function object
        * attributes: All attributes of job that were defined at its creation (including env and vol)
        * profile: Profiling information of the function that ran (if _profile was True)
        * memory: Variety of information about memory usage, swap usage, disk usage, and allocation failures
        * cputime: Amount of time job has spent consuming the cpu (in user and kernel mode)
        * cpu_percentage: Percent of cpu (core) job is currently occupying (in user and kernel mode) (only available as job is processing) 
        * ports: Ports this job has opened (only available as job is processing).
        
        If *info_requested* is None, info_requested will be status, stdout, stderr, and runtime.
        Set *info_requested* to 'all' to obtain info about every possible item                
	    
	    e.g. scicloud.info(42,'stdout') to get standard output of job 42
        """
                
        if not hasattr(jids,'__iter__'):
            jids = [jids]
        
        self._check_jid_type(jids)                
        
        if not info_requested:
            info_requested = ['stdout', 'stderr', 'status', 'runtime']
        
        if not hasattr(info_requested,'__iter__'):
            info_requested = [info_requested]
        
        test = reduce(lambda x, y: x and (isinstance(y,str)), info_requested, True)   
        if not test:
            raise TypeError( 'info_requested must be a valid string value or list of strings' )         

        self._checkOpen()   
        
        cached_info  = self.cacheManager.getCached(jids, info_requested)
        
        #As we only do a single request, we must request all information for any job missing any requested info items
        info_len = len(info_requested)
        needed_jids = filter_xrange_list(lambda jid: len(cached_info.get(jid,[])) != info_len,jids)     
        
        if needed_jids:
            try:
                ainfo_map = self.adapter.jobs_info(jids=needed_jids,info_requested=info_requested)           
            except _catchCloudException, e:
                raise e 
            
            if 'status' not in info_requested:
                cached_statuses = self.cacheManager.getCached(jids, ('status',))
            else:
                cached_statuses = None
        
            can_cache = True
            for jid in ainfo_map.iterkeys():
                #TODO: Correctly cache file_map meta jobs. For now, we must skip caching
                if isinstance(jid, basestring) and 'filemap' in jid:
                    can_cache = False
                    break
                 
            if can_cache:
                for jid, ainfo in ainfo_map.iteritems():
                    status = ainfo.get('status') or cached_statuses.get(jid,{}).get('status',None)
                    if status in self.finished_statuses:
                        self.cacheManager.putCached(jid, **(ainfo))
                       
            cached_info.update(ainfo_map)
        
        for info in cached_info.values():
            fix_time_element(info, 'created')
            fix_time_element(info, 'finished')
            
            depickle_keys = (k for k in info.keys() if k in ['func_object', 'args', 'kwargs'])
            for k in depickle_keys:
                info[k] = pickle.loads(info[k])
            
            
            # convert ports to dictionaries
            if 'ports' in info and hasattr(info['ports'], 'items'):
                for proto, proto_port_dct in info['ports'].items():
                    for port, port_mapping in proto_port_dct.items():
                        port_mapping['port'] = int(port_mapping['port'])
                        del proto_port_dct[port]
                        proto_port_dct[int(port)] = port_mapping
                
        return cached_info 
    
    def kill(self, jids=None):
        """
        Kill any incomplete jobs specified by the integer or iterable *jids*.
        If no arguments are specified (or jids is None), kill every job that has been submitted.
        """
        
        #Internal Note: No caching takes place here, due to rarity of this command
                   
        if jids!=None:
            if not hasattr(jids,'__iter__'):
                jids = [jids]
            
            self._check_jid_type(jids)
            
            if hasattr(jids,'__len__') and len(jids) == 0:
                return

        self._checkOpen()
                
        try:
            self.adapter.jobs_kill(jids)
        except _catchCloudException, e:
            raise e    
    
    def delete(self, jids):
        """
        Remove all data (result, stdout, etc.) related to jobs specified by the 
        integer or iterable *jids* 
        Jobs must have finished already to be deleted.
        
        .. note::
            
            In MP/Simulator mode, this does not delete any datalogs, only in-memory info
        """
        if not hasattr(jids,'__iter__'):
            jids = [jids]
        
        self._check_jid_type(jids)        
        self._checkOpen()
                    
        try:
            self.adapter.jobs_delete(jids)
        except _catchCloudException, e:
            raise e
        finally:
            self.cacheManager.deleteCached(jids)
    
    def map(self, func, *args, **kwargs):
        """
        Map *func* (a callable) over argument sequence(s).
        scicloud.map is meant to mimic a regular map call such as::
            
            map(lambda x,y: x*y, xlist, ylist)
        
        *args* can be any number of iterables. If the iterables are of different
        size, ``None`` is substituted for a missing entries. Note thatunlike call, 
        map does not support passing kwargs into *func*
        
        Map will return an iterable describing integer Job IDentifiers (jids).  Each jid
        corresponds to ``func`` being invoked on one item of the sequence(s) (a 'job').
        In practice, the jids can (and should) be treated as a single jid;
        the returned iterable may be passed directly into status, result, etc.
        
        Using scicloud.result on the returned jids will return a list of job return values 
        (each being the result of applying the function to an item of the argument sequence(s)).
        
        Example::
    
            def mult(x,y):
                return x*y
            jids = scicloud.map(mult, [1,3,5], [2,4,6]) 
    
        This will cause mult3 to be invoked on PiCloud's cluster with x=2
        
        Results::
    
            scicloud.result(jids)
            >> [2,12,30]
        
        The result is [1*2,3*4,5*6]
        
        See online documentation for additional information.
        
        ..note:: If you wish to have a pass a constant value across all map jobs, recommended practice is
            to use `closures <http://marakana.com/bookshelf/python_fundamentals_tutorial/functional_programming.html#_closures>`_
            to close the constant. 
            The `Lambda <http://marakana.com/bookshelf/python_fundamentals_tutorial/functional_programming.html#_anonymous_functions>`_
            keyword is especially elegant::            
                scicloud.map(lambda a, c: foo(a=a, b=b_constant, c=c), a_list, c_list)
                    
        Reserved special *kwargs* (see docs for details) are very similar to call:
        
        * _cores:
            Set number of cores your jobs will utilize. See http://docs.scivm.com/primer.html#choose-a-core-type/
            In addition to having access to more CPU cores, the amount of RAM available will grow linearly.
            Possible values for ``_cores`` depend on what ``_type`` you choose:
            
            * c1: 1                    
            * c2: 1, 2, 4, 8
            * f2: 1, 2, 4, 8, 16                                    
            * m1: 1, 2, 4, 8
            * s1: 1        

        * _depends_on:
            An iterable of jids that represents all jobs that must complete successfully 
            before the jobs created by this map function may be run.
        * _depends_on_errors:
            A string specifying how an error with a jid listed in _depends_on should be handled:       
                 
            * 'abort': Set this job to 'stalled' (Default)
            * 'ignore': Treat an error as satisfying the dependency
        * _env:
            A string specifying a custom environment you wish to run your jobs within.
            See environments overview at http://docs.scivm.com/environment.html                
        * _fast_serialization:
            This keyword can be used to speed up serialization, at the cost of some functionality.
            This affects the serialization of both the arguments and return values            
            Possible values keyword are:
                        
            0. default -- use scicloud module's enhanced serialization and debugging info            
            1. no debug -- Disable all debugging features for arguments            
            2. use cPickle -- Use Python's fast serializer, possibly causing PicklingErrors
            
            ``func`` will always be serialized by the enhanced serializer (with debugging info).                
                    
        * _kill_process:
            Terminate the job's Python interpreter after *func* completes, preventing
            the interpreter from being used by subsequent jobs.  See Technical Overview for more info.                
        * _label: 
            A user-defined string label that is attached to the job. Labels can be
            used to filter when viewing jobs interactively (i.e. on the PiCloud website).
        * _max_runtime:
            Specify the maximum amount of time (in integer minutes) a job can run. If the job runs beyond 
            this time, it will be killed.
        * _os_env_vars:
            List of operating system environment variables that should be copied to PiCloud from your system
            Alternatively a dictionary mapping the environment variables to the desired values.                                  
        * _priority: 
            A positive integer denoting the job's priority. PiCloud tries to run jobs 
            with lower priority numbers before jobs with higher priority numbers.                
        * _profile:
            Set this to True to enable profiling of your code. Profiling information is 
            valuable for debugging, and finding bottlenecks.
            **Warning**: Profiling can slow your job significantly; 10x slowdowns are known to occur
        * _restartable:
            In the rare event of hardware failure while the job is processing, this flag
            indicates whether the job can be restarted. By default, this is True. Jobs
            that modify external state (e.g. adding rows to a database) should set this
            False to avoid potentially duplicate modifications to external state.
        * _type:
            Choose the type of core to use, specified as a string:
            
            * c1: 1 compute unit, 300 MB ram, low I/O (default)                    
            * c2: 2.5 compute units, 800 MB ram, medium I/O
            * f2: 5.5 compute units, 3.75 GB ram, high I/O, hyperthreaded core                                    
            * m1: 3.25 compute units, 8 GB ram, high I/O
            * s1: Up to 2 compute units (variable), 300 MB ram, low I/O, 1 IP per core
                           
            See http://www.scivm.com/pricing/ for pricing information
        * _vol:
            A string or list of strings specifying a volume(s) you wish your job to have access to.
                         
        """
        
        #TODO: Add split/chunking param.. see multiprocessing
        if not callable(func):
            raise TypeError( 'scicloud.map first argument (%s) is not callable'  % (str(func) ))
        
        if len(args) == 0:
            raise ValueError('scicloud.map needs at least 2 arguments')
        
        self._checkOpen()
                
        parameters = self._getJobParameters(func, kwargs)
        if kwargs:
            first_kwd = kwargs.keys()[0]
            raise ValueError('Cannot pass keyword arguments into scicloud.map Unexpected keyword argument %s' % first_kwd)
        
        #typecheck and access correct iterator:
        argIters = []
        cnt = 1
        map_len = 0 #length of this map (i.e. jobs).  None if cannot be determined
        for arg in args:
            try:
                argIters.append(maybe_xrange_iter(arg, coerce = False))
            except AttributeError:
                raise TypeError( 'map argument %d (%s) is not an iterable'  % (cnt,str(arg) ))
            else:
                cnt +=1
                if map_len != None and hasattr(arg, '__len__'):
                    map_len = max(len(arg), map_len)
                else:
                    map_len = None
        
        if map_len > 10000: #hard limit
            raise ValueError('No more than 10,000 arguments can be submitted in scicloud.map')
        
        parameters['map_len'] = map_len 
        
        #[(a,b,c), (1,2,3)] --> [(a,1), (b,2), (c,3)]
        argList = izip_longest(*argIters)
       
        try:
            jids = self.adapter.jobs_map(params=parameters,func=func,mapargs=argList)
        except _catchCloudException, e:
            raise e    
        
        except pickle.PicklingError, e:
            e.args = (e.args[0] + _docmsgpkl,)
            raise e
        
        
        # TODO: ADD SUPPORT FOR CALLBACKS!
        
        return jids
    
class CloudTicketManager(threading.Thread):
    """
    CloudTicketManager is responsible for managing callbacks
    associated with specific jobs/tickets.
    """
    
    scicloud = None
    
    # list of tickets to 
    pending_tickets = None
    
    # condition variable
    cv = None
    
    # dict mapping ticket -> callbacks on success
    _callbacks_on_success = None
    
    # dict mapping ticket -> callbacks on error
    _callbacks_on_error = None
    
    # event indicating if thread should die
    die_event = None
    
    def __init__(self, scicloud):
        threading.Thread.__init__(self)
    
        self.scicloud = scicloud
        self.pending_tickets = []
        self.cv = threading.Condition()
        
        # make this a daemon thread
        self.setDaemon(True)
        
        self._callbacks_on_success = {}
        self._callbacks_on_error = {}
        self.die_event = threading.Event()
        
    def stop(self): 
        self.die_event.set()
        
        with self.cv:    
            self.cv.notify()
        
        self.join() #block until shut down
    
    def add_callbacks(self, ticket, callbacks, type='success'):
        """
        Callbacks should be a function or a list of functions that
        take as its first argument the return value of the scicloud call.
        Callbacks will be called in order from left to right. 
        
        Returns True if callbacks are added, False otherwise. 
        """

        # filters empty lists
        if not callbacks:
            return False
                
        # if callbacks is a function, make it a one item list
        # otherwise, assume it is a list of functions
        if not hasattr(callbacks,'__iter__'):
            callbacks = [callbacks]

        if not reduce(lambda x, y: x and (callable(y)), callbacks, True):
            raise CloudException( 'All callbacks must be callable')
        
        # add to different trigger list depending on type
        if type == 'success':
            self._callbacks_on_success.setdefault(ticket, []).extend(callbacks)
        elif type == 'error':
            self._callbacks_on_error.setdefault(ticket, []).extend(callbacks)
        else:
            raise Exception('Unrecognized callback type.')
    
        # add ticket to watch list
        with self.cv:
            # only add if ticket is not currently present
            if ticket not in self.pending_tickets:
                self.pending_tickets.append(ticket)
                self.cv.notify()
        
        return True
    
    def run(self):
        
        while not self.die_event.isSet():
            with self.cv:
                
                while len(self.pending_tickets) == 0 and not self.die_event.isSet():                                    
                    self.cv.wait()
                
                if self.die_event.isSet():
                    return                
            
            # ASSUMPTION: This is the only thread that will remove tickets
            # TODO: If job does not finish for some time, increase sleep interval
            time.sleep(2.0)
            
            with self.cv:
                
                # get status of tickets
                statuses = self.scicloud.status(self.pending_tickets)
                
                # for each ticket, call callbacks if in a finished state
                for ticket, status in zip(self.pending_tickets, statuses):
                    
                    if status in Cloud.finished_statuses:
                        
                        # call appropriate callback
                        callbacks = self._callbacks_on_success.get(ticket,[]) if status == 'done' else self._callbacks_on_error.get(ticket,[])
         
                        # call callback with function argument if it takes an argument
                        for callback in callbacks:
                            if callback.func_code.co_argcount > 0:
                                callback(ticket)
                            else:
                                callback()
                        
                        # callback triggered -> remove ticket from watch list
                        self.pending_tickets.remove(ticket)
                    
"""hack to handle S3 glitches with scivm bucket"""
if Cloud.running_on_scicloud():
    import socket
    socket.setdefaulttimeout(60)



"""
running_on_scicloud version of bucket interface that accesses /bucket mount point to do bucket operations
Do NOT directly use this in your own code!
"""
from __future__ import with_statement
from __future__ import absolute_import

"""
TODO:
Consider having connection straight to the boss to sign requests to get/put
""" 

"""Currently not supported:

-getf with byte ranges
-modifying content-type/encoding (uses s3fs default)
-list/iterlist (fast enough to go to webserver / semantically nontrivial to use s3fs w/ limited keys)
-remove_prefix - s3fs lacks support for batch delete
-make_public & friends - s3fs lacks support
-info - s3fs lacks support for metadata requests
-sync_*: Look into how to do this without bringing the whole file in; see rsync trickery of s3fs
    Right now sync just falls back to always put/get file.
        In general sync is less practical server-side
"""

"""
Copyright (c) 2013 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@piscicloud.com

The scicloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see 
http://www.gnu.org/licenses/lgpl-2.1.html    
"""


import errno
import os
import time
import logging
from .scicloud import CloudException, CloudTimeoutError

scicloudLog = logging.getLogger('Cloud.bucket')

# base implementation of bucket 
from . import bucket as base_bucket

# auxilerary functions that don't directly touch website
from .bucket import _compute_md5, _get_effective_obj_path, _putf, _ready_file_path, \
                    S3_URL, TruncatableList, builtin_list

# fallback to default implementations
from .bucket import list, iterlist, remove_prefix, make_public, public_url_folder, \
                    is_public, make_private, info, get_md5, _get_md5

# Error codes used by bucket:
filenametoolong = 491
filenotfound = 492

# human readable messages
hrm = {
       filenametoolong: 'The specified filename was too long.',
       filenotfound : 'The specified object was not found.'
       }


def _emulate_apiexception(error_code):
    error_hrm = hrm[error_code]
    return CloudException(error_hrm, jid=None, status=error_hrm)


bucket_basepath = '/bucket'

def _retry_op(func, excp_class=Exception, _num_retries=2):
    retry_cnt = 0
    while retry_cnt <= _num_retries:
        try:
            retval = func()
        except excp_class, e:
            if retry_cnt == _num_retries:
                raise
            else:
                scicloudLog.warning('Received exception %s:%s. Retrying', 
                                 type(e), str(e), exc_info=True)
                time.sleep(0.2)
                continue
        else:
            return retval    

def _get_bucket_path(obj_path, prefix):
    full_obj_path = _get_effective_obj_path(obj_path, prefix)
    return os.path.join(bucket_basepath, full_obj_path)

def putf(f, obj_path, prefix=None, _buffersize = 16384):
    """
    helper for putf.
    Does not support content_type/content_encoding
    """
            
    dest_path = _get_bucket_path(obj_path, prefix)
    
    fsize = 0 # file size. may not be computable 
    
    if isinstance(f, basestring):
        from cStringIO import StringIO        
        f = StringIO(f)
    
    try:
        start_loc = f.tell()
        f.seek(0,2)        
        fsize = f.tell() - start_loc
        f.seek(start_loc)
    except IOError:  
        raise IOError('File object is not seekable. Cannot transmit')

    if fsize > 5000000000:
        raise ValueError('Cannot store bucket objects larger than 5GB on scicloud.bucket')
    
    if fsize == 0:
        raise ValueError('Cannot store empty bucket objects')
    
    _ready_file_path(dest_path)
    
    def writer():
        with open(dest_path, 'wb') as destf:
            
            data = f.read(_buffersize)
            while data:         
                destf.write(data)
                data = f.read(_buffersize)
    
    try:
        _retry_op(writer,EnvironmentError)
    finally:
        f.close()
        
def put(file_path, obj_path=None, prefix=None):
    
    if obj_path is None:
        obj_path = os.path.basename(file_path)
    elif not obj_path:
        raise ValueError('Cannot upload bucket object with obj_path "%s"' % obj_path)

    f = open(file_path, 'rb')        
    putf(f, obj_path, prefix)
    
# Serverside md5 computation will take too long to be worthwhile, just always do a put
#  TODO: With boss signing, md5 checks may be doiable
sync_to_scicloud = put
    
def exists(obj_path, prefix=None):
    bucket_path = _get_bucket_path(obj_path, prefix)
    return _retry_op(lambda: os.path.exists(bucket_path), EnvironmentError)

def remove(obj_paths, prefix=None):
    """Removes object(s) named ``effective_obj_paths`` from PiCloud bucket
    
    obj_paths can be a single object or a list of objects
    """    
    
    if not hasattr(obj_paths, '__iter__'):
        obj_paths = [obj_paths]
    
    if len(obj_paths) > 10: # if batch, use webserver batch remove
        return base_bucket.remove(obj_paths, prefix)
        
    for obj_path in obj_paths:
        bucket_path = _get_bucket_path(obj_path, prefix)
                
        try:
            os.remove(bucket_path)
        except OSError, oe:
            if oe.errno != errno.ENOENT:
                raise

def get(obj_path, file_path=None, prefix=None, start_byte=0, end_byte=None, _retries=1):
    """Similar to regular get, but bypass md5 checking"""
    
    if start_byte or end_byte:
        # cannot handle ranges with s3fs without pulling entire file in
        # fall back to default implementation        
        return base_bucket.get(obj_path, file_path, prefix, start_byte, end_byte, _retries)
    
    file_path = _ready_file_path(file_path, obj_path)    

    # with s3fs, we are effectively doing a copy
    scicloud_file = getf(obj_path, prefix, 0, None)
    
    chunk_size = 16384
    
    f = open(file_path, 'wb')
    
    while True:
        data = scicloud_file.read(chunk_size)
        if not data:
            break
        f.write(data)
    
    f.close()
    
# always get on server (md5 check takes too long)
sync_from_scicloud = get        

"""get logic"""

# redefined 
class CloudBucketObject(file):
    """A CloudBucketObject provides a file-like interface to the contents of an object
    On the server-side this is just a file subclass made compatible with client-side CloudBucketObject"""
    
    __cached_md5 = None
       
    @property
    def md5(self):
        # interferes with read!
        if self.__cached_md5 == None:
            self.__cached_md5, _, _ = _compute_md5(self)
        return self.__cached_md5

    def filesize(self):
        # not threadsafe with respect to read() 
        start_loc = self.tell()
        self.seek(0,2)
        size = self.tell()
        self.seek(start_loc)
        
        return size
    
    def sizeofchunk(self):
        return self.filesize()
    
    def end(self):
        return self.filesize()

def getf(obj_path, prefix=None, start_byte=0, end_byte=None):
    """
    Retrieve the object referenced by ``effective_obj_path`` from PiCloud.
    Return value is a CloudBucketObject (file-like object) that can be read() to 
    retrieve the object's contents 

    An optional byte_range can be specified using ``start_byte`` and ``end_byte``, 
    where only the data between ``start_byte`` and ``end_byte`` is returned and made 
    accessible to the CloudBucketObject.  The returned CloudBucketObject.tell() will 
    initialized to ``start_byte``.
    
    An ``end_byte`` of None or exceeding file size is interpreted as a request to retrieve to end of file.
    """    
    
    if start_byte or end_byte:
        # cannot handle ranges with s3fs without pulling entire file in
        # fall back to default implementation        
        return base_bucket.getf(obj_path, prefix, start_byte, end_byte)
    
    
    bucket_path = _get_bucket_path(obj_path, prefix)
    
    try:
        return CloudBucketObject(bucket_path, 'rb')
    except OSError, e:
        if e.errno == errno.ENOENT:
            raise _emulate_apiexception(filenotfound)
        else:
            raise

def mpsafe_get(obj_path, file_path=None, prefix=None, start_byte=0, end_byte=None, 
               timeout=None, do_sync=False):
    """Multiprocessing-safe variant of get.
    If do_sync is false:
        Download file if and only if file_path does not exist. (atomically checked)
    If do_sync is true:
        Download file if and only if copy at file_path does not match remote copy (atomically checked)
    Regardless, If another process is downloading, wait until process has finished
    
    If timeout is reached, CloudTimeoutException is raised
    """
    
    # Exact same code as base bucket: Here only so it references different get function
    
    from .util.cloghandler import portalocker
    
    file_path = _ready_file_path(file_path, obj_path)    
    
    # lock the file itself
    try:
        #fd = os.open(file_path, os.O_CREAT|os.O_RDONLY)
        fobj = open(file_path,'a')
    except OSError as oe:
        if oe.errno != errno.EEXIST:
            raise
        
    expire_time = timeout + time.time() if timeout else None 
    while True:
        try:
            portalocker.lock(fobj, (portalocker.LOCK_EX | portalocker.LOCK_NB))
        except portalocker.LockException as le:
            if le.args[0] != portalocker.LockException.LOCK_FAILED:
                raise
            if expire_time and time.time() > expire_time:
                raise CloudTimeoutError('could not acquire file lock within timeout')
            
            time.sleep(0.2) # delay
        else:
            break
            
    # We now have the lock; determine what to do based on file size
    # Zero file size is special as buckets cannot contained zero sized files
    # If file currently is zero-sized, we are the first writer: download it
    #   Else, someone else got to it first    
    fsize = os.path.getsize(file_path)
    if not fsize: # download
        get(obj_path, file_path, prefix, start_byte, end_byte)
    elif do_sync:
        sync_from_scicloud(obj_path, file_path, prefix)         
    
    fobj.close()  # releases lock



"""
DEPRECATED: Please use scicloud.bucket

For managing files on Science VM's S3 store.

.. note::

    This module cannot be used to access files stored on a job's mounted file system
"""

# TODO: Support files larger than 5 gb

from __future__ import with_statement
from __future__ import absolute_import
"""
Copyright (c) 2011 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@piscicloud.com

The scicloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see 
http://www.gnu.org/licenses/lgpl-2.1.html    
"""

"""User beware: list is defined in this module; list will not map to builtin list!"""

import os
import sys
import logging
import time
import socket
import errno
import random
import urllib2

__httpConnection = None
__url = None
#__query_lock = threading.Lock() #lock on http adapter updates

from .transport.adapter import SerializingAdapter
from .transport.network import HttpConnection
from .util import  min_args, max_args
from .util.zip_packer import Packer
from .scicloud import CloudException
from scicloud import _getscicloudnetconnection, _getscicloud

scicloudLog = logging.getLogger('Cloud.files')

_file_new_query = 'file/new/'
_file_put_query = 'file/put/'
_file_list_query = 'file/list/'
_file_get_query = 'file/get/'
_file_exists_query = 'file/exists/'
_file_md5_query = 'file/md5/'
_file_delete_query = 'file/delete/'
_filemap_job_query = 'job/filemap/'
_default_chunk_size = 200

"""
This module utilizes the scicloud object extensively
The functions can be viewed as instance methods of the Cloud (hence accessing of protected variables)
"""

def _post(conn, url, post_values, headers={}):
    """Use HttpConnection *conn* to issue a post request at *url* with values *post_values*"""
    
    #remove UNICODE from addresses
    url = url.decode('ascii', 'replace').encode('ascii', 'replace')
    
    if 'success_action_redirect' in headers:
        headers['success_action_redirect'] = headers['success_action_redirect'].decode('ascii', 'replace').encode('ascii', 'replace')
    if post_values and 'success_action_redirect' in post_values:
        post_values['success_action_redirect'] = post_values['success_action_redirect'].decode('ascii', 'replace').encode('ascii', 'replace')
    
    scicloudLog.debug('post url %s with post_values=%s. headers=%s' % (url, post_values, headers))
    response =  conn.post(url, post_values, headers, use_gzip=False)
    
    return response

def _aws_retryable_post(conn, url, post_values, headers={}):
    """Wraps _post with ability to retry"""
    retry_attempts = conn.retry_attempts
    attempt = 0
    while attempt <= retry_attempts:    
        try:
            return _post(conn, url, post_values, headers)
        except Exception, e:
            if isinstance(e, urllib2.HTTPError) and 200 < e.code < 300: # python 2.5 bug
                return e
            
            attempt += 1            
            if attempt > retry_attempts:
                scicloudLog.exception('_aws_retryable_post: Cannot connect to AWS')
                raise 
            scicloudLog.warn('_aws_retryable_post: Problem connecting to AWS. Retrying. \nError is %s' % str(e))
            c = attempt -1 
            if (isinstance(e, socket.error) and getattr(e, 'errno', e.args[0]) == errno.ECONNREFUSED) or \
                (isinstance(e, urllib2.HTTPError) and e.code in [500, 503]):
                # guarantee at least a 3 second sleep on a connection refused error/500
                time.sleep(min(60, 3 + (1 << c) *random.random()))
            else:
                time.sleep(min(30, (1 << c) *random.random()))
            continue
            
        

class CloudFile(object):
    """A CloudFile represents a file stored on PiCloud as a readonly file-like stream.
    Seeking is not available."""
    
    __http_response = None

    __file_size = None
    __start_byte = None
    __end_byte = None
    __pos = None
    
    def __init__(self, http_response, file_size, start_byte=0, end_byte=None):
        self.__http_response = http_response
        
        self.__file_size = file_size
        self.__start_byte = start_byte
        self.__pos = self.__start_byte
        if end_byte:
            self.__end_byte = min(end_byte, file_size)
        else:
            self.__end_byte = file_size
                
    def __iter__(self):
        return self

    def close(self):        
        """
        Close the file, blocking further I/O operations
        """
        return self.__http_response.close()
    
    @property
    def md5(self):
        """The md5 checksum of file contents"""
        return self.__http_response.headers['etag'].strip('"')
    
    def next(self):
        data = self.__http_response.next()
        self.__pos += len(data)
        return data
    
    def read(self, size=-1):
        """
        read([size]) -> read at most size bytes, returned as a string.
       
        If the size argument is negative or omitted, read until EOF is reached.
        """
        data = self.__http_response.read(size)
        self.__pos += len(data)
        return data
        
    def readline(self, size=-1):
        """
        readline([size]) -> next line from the file, as a string.
       
        Retain newline.  A non-negative size argument limits the maximum
        number of bytes to return (an incomplete line may be returned then).
        Return an empty string at EOF.
        """
        line = self.__http_response.readline(size)
        self.__pos += len(line)
        return line
        
    def readlines(self, sizehint=0):
        """
        readlines([size]) -> list of strings, each a line from the file.
       
        Call readline() repeatedly and return a list of the lines so read.
        The optional size argument, if given, is an approximate bound on the
        total number of bytes in the lines returned.
        """
        lines = self.__http_response.readlines(sizehint)
        self.__pos += sum([len(line) for line in lines])
        return lines

    def filesize(self):
        return self.__file_size
    
    def sizeofchunk(self):
        return self.__start_byte - self.__end_byte
    
    def tell(self):
        '''Returns current file position as an integer'''
        return self.__pos

    def byte_range(self):
        return self.__byte_range

    def end(self):
        return self.__end_byte
 
def _compute_md5(f, buffersize = 8192):
    """Computes the md5 hash of file-like object
    f must have seek ability to perform this operation
    
    buffersize controls how much of file is read at once"""
    
    try:
        start_loc = f.tell()
        f.seek(start_loc)  #test seek
    except (AttributeError, IOError), e:
        raise IOError('%s is not seekable. Cannot compute MD5 hash. Exact error is %s' % (str(f), str(e)))
    
    try:
        from hashlib import md5
    except ImportError:
        from md5 import md5
    
    m = md5()    
    s = f.read(buffersize)
    while s:
        m.update(s)
        s = f.read(buffersize)
    
    f.seek(start_loc)
    hex_md5 = m.hexdigest()
    return hex_md5        

def put(source, name=None):
    """Transfers the file specified by ``file_path`` to PiCloud. The file
    can be retrieved later using the get function.    
    
    If ``name`` is specified, the file will be stored as name on PiCloud.
    Otherwise it will be stored as the basename of file_path.
    
    Example::    
    
        scicloud.files.put('data/names.txt') 
    
    This will transfer the file from the local path 'data/names.txt'
    to PiCloud and store it as 'names.txt'.
    It can be later retrieved via scicloud.files.get('names.txt')"""

    if not name:
        name = os.path.basename(source)
    
    # open the requested file in binary mode (relevant in windows)
    f = open(source, 'rb')
    
    putf(f, name)


def putf(f, name):
    """Similar to put.
    putf, however, accepts a file object (file, StringIO, etc.) ``f`` instead of a file_path.
    
    .. note::
        
        ``f`` is not rewound. f.read() from current position will be placed on PiCloud
    
    .. warning:: 
    
        If the file object does not correspond to an actual file on disk,
        it will be read entirely into memory before being transferred to PiCloud."""
    
    if '../..' in name:
        raise ValueError('"../.." cannot be in name')
    
    fsize = 0 # file size. may not be computable 
    
    if isinstance(f, basestring):
        fsize = len(f)                        
        from cStringIO import StringIO        
        f = StringIO(f)
    else:
        try:
            fsize = os.fstat(f.fileno()).st_size
        except (AttributeError, OSError):
            pass
    
    if fsize > 5000000000:
        raise ValueError('Cannot store files larger than 5GB on scicloud.files')
    
    conn = _getscicloudnetconnection()         
    
    try:
        # get a file ticket
        resp = conn.send_request(_file_new_query, {'name': name})
        ticket = resp['ticket']
        params = resp['params']
        
        url = params['action']
        
        # set file in ticket
        ticket['file'] = f
        
        # post file using information in ticket
        ticket['key'] = str(ticket['key'])
        resp =  _aws_retryable_post(conn, url, ticket)
        resp.read()
        
    finally:
        f.close()

def sync_to_scicloud(source, name=None):
    """Upload file if it has changed.
    
    scicloud.files.put(source,name) 
    only if contents of local file (specified by *source*)
    differ from those on PiCloud (specified by *name* or basename(*source*))
    (or if file does not exist on PiCloud)"""
    if not name:
        name = os.path.basename(source)
    
    # open the requested file in binary mode (relevant in windows)
    f = open(source, 'rb')
    local_md5 = _compute_md5(f)
    try:
        remote_md5 = get_md5(name, log_missing_file_error = False)
    except CloudException: #file not found
        remote_md5 = ''
    
    do_update = remote_md5 != local_md5
    scicloudLog.debug('remote_md5=%s. local_md5=%s. uploading? %s',
                   remote_md5, local_md5, do_update
                   )
    if do_update:
        putf(f, name)
 
def list():
    """List all files stored on PiCloud."""
    
    conn = _getscicloudnetconnection()

    resp = conn.send_request(_file_list_query, {})
    files = resp['files']

    return files

def _file_info(name):
    """
    get information about name
    """
    conn = _getscicloudnetconnection()
    
    resp = conn.send_request(_file_exists_query, {'name':name})
    return resp

def exists(name):
    """Check if a file named ``name`` is stored on PiCloud."""
    conn = _getscicloudnetconnection()
        
    resp = conn.send_request(_file_exists_query, {'name': name})
    exists = resp['exists']
    return exists

def get_md5(name, log_missing_file_error = True):
    """Return the md5 checksum of the file named ``name`` stored on PiCloud"""
    conn = _getscicloudnetconnection()
    resp = conn.send_request(_file_md5_query, {'name': name},
                             log_scicloud_excp = log_missing_file_error)
    md5sum = resp['md5sum']
    return md5sum
    
def delete(name):
    """Deletes the file named ``name`` from PiCloud."""
    conn = _getscicloudnetconnection()

    resp = conn.send_request(_file_delete_query, {'name': name})
    deleted = resp['deleted']
    return deleted
    
def get(name, destination=None, start_byte=0, end_byte=None):
    """
    Retrieves the file named by ``name`` from PiCloud and stores it to ``destination``.
        
    Example::    
    
        scicloud.files.get('names.txt','data/names.txt') 
    
    This will retrieve the 'names.txt' file on PiCloud and save it locally to 
    'data/names.txt'. 

    If destination is None, destination will be name
    
    An optional byte_range can be specified using *start_byte*, *end_byte*,
    where only the data between *start_byte* and *end_byte* is returned. 
    If end_byte exceeds the size of the file, the contents from *start_byte* to end of file returned.
    
    An end_byte of None or exceeding file size is interpreted as end of file
    """
    
    if not destination:
        destination = name
        
    scicloud_file = getf(name, start_byte, end_byte)
    
    chunk_size = 64000
    f = open(destination, 'wb')
    
    while True:
        data = scicloud_file.read(chunk_size)
        if not data:
            break
        f.write(data)
    
    f.close()

    
def sync_from_scicloud(name, destination=None):
    """Download file if it has changed.
    
    scicloud.files.get(name,destination) 
    only if contents of local file (specified by *destination* or basename(*name))
    differ from those on PiCloud (specified by *name*) 
    (or *destination* does not exist locally)"""
    
    if not destination:
        destination = name
    
    if not os.path.exists(destination):
        do_update = True
    else:
        f = open(destination, 'rb')
        local_md5 = _compute_md5(f)
        f.close()
        try:
            remote_md5 = get_md5(name, log_missing_file_error=False)
        except CloudException: #file not found
            remote_md5 = ''
    
        do_update = remote_md5 != local_md5
        scicloudLog.debug('remote_md5=%s. local_md5=%s. downloading? %s',
                       remote_md5, local_md5, do_update)
    if do_update:
        get(name, destination)

def getf(name, start_byte=0, end_byte=None):
    """
    Retrieves the file named by ``name`` from PiCloud.
    Return value is a CloudFile (file-like object) that can be read() to retrieve the file's contents 

    A range can be specified through *start_byte* and *end_byte*, where only the data between those two offsets
    will be accessable in the CloudFile.  If start_byte is set, the returned CloudFile.tell() will be start_byte
    
    An end_byte of None or exceeding file size is interpretted as end of file
    """    
    
    conn = _getscicloudnetconnection()

    resp = conn.send_request(_file_get_query, {'name': name})
    
    ticket = resp['ticket']
    params = resp['params']
    file_size = params['size']
    
    if not start_byte:
        start_byte=0
    if file_size and (not end_byte or end_byte > file_size):
        end_byte = file_size

    if not isinstance(start_byte, (int, long)):
        raise TypeError('start_byte must be an integer')
    
    if end_byte and not isinstance(end_byte, (int, long)):
        raise TypeError('end_byte must be an integer')

    if end_byte:
        ticket['Range'] = 'bytes=%s-%s' % tuple(  [start_byte, end_byte]  )

    resp =  _aws_retryable_post(conn, params['action'], None, ticket)
    
    scicloud_file = CloudFile( resp, file_size, start_byte, end_byte )
    
    return scicloud_file


def default_record_reader(delimiter):

    def def_record_reader(filesplit_obj, end_byte):

        start = filesplit_obj.tell()
        end = end_byte

        if start==0:
            skipfirstline = False       # we are reading the head of the file
        else:
            skipfirstline = True        # we are reading the middle of a chunk returned in another job
 
        to_be_searched_buffer = ''      # the string that will be searched for the delimiter
        record_buffer = ''              # Record that has been read so far 
        record_start = start            # pointer to the beginning of the record 
        crossed_end_of_chunk = False
        
        while filesplit_obj.tell() < filesplit_obj.filesize():
            
            to_be_searched_buffer = filesplit_obj.read(1024)

            while True:
                
                partition = to_be_searched_buffer.partition(delimiter)
                
                if len(partition[1])==0: # delimiter not found
                    record_buffer = record_buffer + to_be_searched_buffer
                    record_start = record_start + len(to_be_searched_buffer)
                
                    if filesplit_obj.tell()==filesplit_obj.filesize():  # are we at EOF?
                        if not skipfirstline:
                            yield record_buffer
                    
                    break
                
                else:                   # delimiter has been found
                    if skipfirstline:
                        skipfirstline = False
                    else:
                        yield record_buffer + partition[0]
                        
                    index = len(partition[0])
                    if end < (record_start + index):
                        crossed_end_of_chunk = True
                        break
                    
                    to_be_searched_buffer = partition[2]
                    record_buffer=''
                    record_start = record_start + index + 1
                        
            if crossed_end_of_chunk:
                break
         
        filesplit_obj.close()
        
    return def_record_reader


def _validate_arguments(func_arg, param_name):
    """Validate that certain map parameters are callables that take 1 argument"""
    if not callable(func_arg):
        raise TypeError( '%s argument (%s) must be callable'  % (param_name, str(func_arg)) )
    
    try:
        max_arg = max_args(func_arg)
        min_arg = min_args(func_arg)
    except TypeError:
        pass #type can't be sanity checked.. let it through
    else:
        if max_arg < 1 or min_arg > 1:
            raise TypeError( '%s argument (%s) must accept one parameter'  % (param_name, str(func_arg)) )            

def _validate_rr_arguments(func_arg, param_name):
    """Validate that certain map parameters are callables that take 1 argument"""
    if not callable(func_arg):
        raise TypeError( '%s argument (%s) must be callable'  % (param_name, str(func_arg)) )
    
    try:
        max_arg = max_args(func_arg)
        min_arg = min_args(func_arg)
    except TypeError:
        pass #type can't be sanity checked.. let it through
    else:
        if max_arg < 2 or min_arg > 2:
            raise TypeError( '%s argument (%s) must accept two parameters'  % (param_name, str(func_arg)) )            

def _mapper_combiner_wrapper(mapper, name, file_size, record_reader, combiner):
    
    def inner(server_mapper):
        return server_mapper(mapper, name, file_size, record_reader, combiner)
    
    return inner

def _reducer_wrapper(reducer):
    
    def inner(server_reducer):
        return server_reducer(reducer)
    
    return inner

def map(name, mapper, chunk_size=None, record_reader=None, combiner=None, reducer=None, **kwargs):
    """
    With map, you can process a file stored in scicloud.files in parallel. The
    parallelism is achieved by dividing the file specified by *name* into
    chunks of size *chunk_size* (bytes). Each chunk is assigned a sub job. The
    sub job in turn processes just that chunk, allowing for the entire file to
    be processed by as many cores in parallel as there are chunks. We will call
    this type of sub job a "mapper sub job".
    
    If chunk_size is None, it will be automatically set to 1/10th of the size
    of the file.
    
    Map will return a single job IDentifier (jid). The sub jobs that comprise it
    do not have identifiers and, therefore, cannot be accessed directly.
    scicloud.info(jid), however, will show you information for relevant sub jobs.
    
    By default, each chunk is split into records (of 0 or more characters) using
    newlines as delimiters. If *record_reader* is specified as a string, each
    chunk is split into records using that as the delimiter.
    
    In the event a record spans across two chunks, it is guaranteed that a mapper
    will only be called once on the full record. In other words, we've made sure
    it works correctly.
    
    *mapper* is a function that takes a single argument, a record, and should
    return an iterable of values (a generator). In the simplest case, it can
    return a generator that yields only one value.
    
    Example::
    
        def mapper(record):
            yield record
    
    When no *combiner* or *reducer* is specified, the return value of the
    scicloud.files.map job will be roughly equivalent to::
            
            map(mapper, record_reader(file_contents))
    
    A *reducer* is a function that takes in an iterable of values and returns an 
    iterable of values.  The iterable parameter iterates through all the values 
    returned by all the mapper(record) calls. When the reducer is specified,
    *reducer* will result in the creation of one additional sub job. The reducer
    sub job grabs the results of each mapper sub job (iterators), combines them
    into a single iterator, and then passes that iterator into your *reducer*
    function. The return value of the scicloud.files.map job will be the iterator
    returned by the *reducer*.
    
    A *combiner*, like a *reducer*, takes in an iterable of values and returns an
    iterable of values. The difference is that the *combiner* is run in each
    mapper sub job, and each one only takes in values that were produced from the
    associated chunk. If a *reducer* is also specified, then the reducer sub job
    grabs the results of each *combiner* run in each mapper sub job.
    
    Example for counting the number of words in a document::
    
        def wordcount_mapper(record):
            yield len(record.split(' '))
            
        def wordcount_reducer(wordcounts):
            yield sum(wordcounts)
            
        jid = scicloud.files.map('example_document', wordcount_mapper, reducer=wordcount_reducer)
        
    Result::
        scicloud.result(jid)
            >> [# of words]
    
    For advanced users, *record_reader* can also be specified as a function that
    takes in a file-like object (has methods read(), tell(), and seek()), and
    the end_byte for the current chunk. The *record_reader* should return an
    iterable of records.  See default_record_reader for an example.
    
    Additional information exists on our blog and online documentation.
        
        Reserved special *kwargs* (see docs for details):
        
        * _cores:
            Set number of cores your job will utilize. See http://docs.scivm.com
            In addition to having access to more cores, you will have _cores*RAM[_type] where _type is the _type you select
            Possible values depend on what _type you choose:
            
            'c1': 1
            'c2': 1, 2, 4, 8
            'f2': 1, 2, 4, 8, 16
            'm1': 1, 2
            's1': 1        
        * _depends_on:
            An iterable of jids that represents all jobs that must complete successfully 
            before any jobs created by this map function may be run.
        * _depends_on_errors:
            A string specifying how an error with a jid listed in _depends_on should be handled.
            'abort': Set this job to 'stalled'  (Default)
            'ignore': Treat an error as satisfying the dependency            
        * _env:
            A string specifying a custom environment you wish to run your jobs within.
            See environments overview at 
            http://docs.scivm.com
        * _fast_serialization:
            This keyword can be used to speed up serialization, at the cost of some functionality.
            This affects the serialization of both the map arguments and return values
            The map function will always be serialized by the enhanced serializer, with debugging features.
            Possible values keyword are:
                        
            0. default -- use scicloud module's enhanced serialization and debugging info            
            1. no debug -- Disable all debugging features for arguments            
            2. use cPickle -- Use python's fast serializer, possibly causing PicklingErrors                
        * _kill_process:
                Terminate the Python interpreter *func* runs in after *func* completes, preventing
                the interpreter from being used by subsequent jobs.  See Technical Overview for more info.                            
        * _label: 
            A user-defined string label that is attached to the created jobs. 
            Labels can be used to filter when viewing jobs interactively (i.e.
            on the PiCloud website).        
        * _max_runtime:
            Specify the maximum amount of time (in integer minutes) a job can run. If job runs beyond 
            this time, it will be killed.                     
        * _priority: 
                A positive integer denoting the job's priority. PiCloud tries to run jobs 
                with lower priority numbers before jobs with higher priority numbers.            
        * _profile:
                Set this to True to enable profiling of your code. Profiling information is 
                valuable for debugging, but may slow down your job.
        * _restartable:
                In the very rare event of hardware failure, this flag indicates that the job
                can be restarted if the failure happened in the middle of the job.
                By default, this is true. This should be unset if the job has external state
                (e.g. it modifies a database entry)
        * _type:
                Select the type of compute resources to use.  PiCloud supports four types,
                specified as strings:
                
                'c1'
                    1 compute unit, 300 MB ram, low I/O (default)                    
                'c2'
                    2.5 compute units, 800 MB ram, medium I/O                    
                'm1'                    
                    3.25 compute units, 8 GB ram, high I/O
                's1'
                    variable compute units (2 cu max), 300 MB ram, low I/O, 1 IP per core                    
                                    
                See http://www.scivm.com/pricing/ for pricing information
    """
    
    scicloud_obj = _getscicloud()
    params = scicloud_obj._getJobParameters(mapper, kwargs)    # takes care of kwargs
    
    file_details = _file_info(name)
    if not file_details['exists']:
        raise ValueError('file does not exist on the scicloud, or is not yet ready to be accessed')
    file_size = int( file_details['size'] )
    params['file_name'] = name
    
    
    # chunk_size
    if chunk_size:
        if chunk_size==0:
            raise Exception('the chunk_size should be a non zero integer value')
        if not isinstance(chunk_size, (int, long)):
            raise Exception('the chunk_size should be a non zero integer value')        
        params['chunk_size'] = chunk_size
            
    
    # mapper
    _validate_arguments(mapper, 'mapper')
    
    # record_reader
    if not record_reader:
        record_reader = default_record_reader('\n')
    else:
        if isinstance(record_reader, basestring):
            record_reader = default_record_reader(record_reader)
        else:
            _validate_rr_arguments(record_reader, 'record_reader')
    
    # combiner
    if not combiner:
        def combiner(it):
            for x in it:
                yield x
    else:
        _validate_arguments(combiner, 'combiner')


    func_to_be_sent = _mapper_combiner_wrapper(mapper, name, file_size, record_reader, combiner)
    
    sfunc, sarg, logprefix, logcnt = scicloud_obj.adapter.scicloud_serialize( func_to_be_sent, 
                                                                    params['fast_serialization'], 
                                                                    [], 
                                                                    logprefix='mapreduce.' )
    
    data = Packer()
    data.add(sfunc)
    params['data'] = data.finish()
    
    # validate reducer & serialize reducer
    if reducer:
        _validate_arguments(reducer, 'reducer')
        reducer = _reducer_wrapper(reducer)
        s_reducer, red_sarg, red_logprefix, red_logcnt = scicloud_obj.adapter.scicloud_serialize( reducer, params['fast_serialization'], [], logprefix='mapreduce.reducer.' )
        data_red = Packer()
        data_red.add(s_reducer)
        params['data_red'] = data_red.finish()
        
    conn = _getscicloudnetconnection()
    conn._update_params(params)
    scicloud_obj.adapter.dep_snapshot()
    
    resp = conn.send_request(_filemap_job_query, params)
    
    return resp['jids']



#!/usr/bin/python
"""
Entry Point for the Science VM Command-Line Interface (CLI)
"""
# since this module sits in the scicloud package, we use absolute_import
# so that we can easily import the top-level package, rather than the
# scicloud module inside the scicloud package
from __future__ import absolute_import
"""
Copyright (c) 2011 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@piscicloud.com

The scicloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see 
http://www.gnu.org/licenses/lgpl-2.1.html    
"""

import sys
import logging
import traceback

try:
    import json
except:
    # Python 2.5 compatibility
    import simplejson as json

import scicloud
from UserDict import DictMixin

from . import argparsers
from .util import list_of_dicts_printer, dict_printer, list_printer,\
    key_val_printer, volume_ls_printer, scicloud_info_printer,\
    scicloud_result_printer, scicloud_result_json_printer, bucket_info_printer,\
    no_newline_printer
from .setup_machine import setup_machine
from . import functions


def main(args=None):
    """Entry point for PiCloud CLI. If *args* is None, it is assumed that main()
    was called from the command-line, and sys.argv is used."""
    
    args = args or sys.argv[1:]
        
    # special case: we want to provide the full help information
    # if the cli is run with no arguments
    if len(args) == 0:
        argparsers.scivm_parser.print_help()
        sys.exit(1)
    
    # special case: if --version is specified at all, print it out
    if '--version' in args:
        print 'scicloud %s' % scicloud.__version__
        print 'running under python %s' % sys.version
        sys.exit(0)
        
    # parse_args is an object whose attributes are populated by the parsed args
    parsed_args = argparsers.scivm_parser.parse_args(args)
    module_name = getattr(parsed_args, '_module', '')
    command_name = getattr(parsed_args, '_command', '')
    function_name = module_name + ('.%s' % command_name if command_name else '')
        
    if function_name != 'setup':
        # using showhidden under setup will cause config to be flushed with hidden variables
        scicloud.config._showhidden()
        scicloud.config.verbose = parsed_args._verbose
        # suppress log messages
        scicloud.config.print_log_level = logging.getLevelName(logging.CRITICAL)

    if parsed_args._api_key:
        scicloud.config.api_key = parsed_args._api_key
    if parsed_args._api_secretkey:
        scicloud.config.api_secretkey = parsed_args._api_secretkey
    if parsed_args._simulate:
        scicloud.config.use_simulator = parsed_args._simulate     
    scicloud.config.commit()
        
    # we take the attributes from the parsed_args object and pass them in
    # as **kwargs to the appropriate function. attributes with underscores
    # are special, and thus we filter them out.
    kwargs = dict([(k, v) for k,v in parsed_args._get_kwargs() if not k.startswith('_')])
                
    # handle post-op 
    for key, value in kwargs.items():
        if callable(value):
            kwargs[key] = value(**kwargs)

    
    # we keep function_mapping and printer_mapping here to prevent
    # circular imports
    
    # maps the output of the parser to what function should be called
    function_mapping = {'setup': setup_machine,
                        'exec': functions.execute,
                        'mapexec': functions.execute_map,
                        'status': functions.status,
                        'join': functions.join,
                        'result': functions.result,
                        'info': functions.info,
                        'kill': functions.kill,                        
                        'delete': functions.delete,
                        'ssh-info': scicloud.shortcuts.ssh.get_ssh_info,
                        'ssh' : functions.ssh,
                        'exec-shell' : functions.exec_shell,
                        'rest.publish': functions.rest_publish, # move to rest?
                        'rest.remove' : scicloud.rest.remove,
                        'rest.list' : scicloud.rest.list,
                        'rest.info' : scicloud.rest.info,
                        'rest.invoke' : functions.rest_invoke,
                        'rest.mapinvoke' : functions.rest_invoke_map,                                                
                        'files.get': scicloud.files.get,
                        'files.put': scicloud.files.put,
                        'files.list': scicloud.files.list,
                        'files.delete': scicloud.files.delete,
                        'files.get-md5': scicloud.files.get_md5,
                        'files.sync-from-scicloud': scicloud.files.sync_from_scicloud,
                        'files.sync-to-scicloud': scicloud.files.sync_to_scicloud,
                        'bucket.get': scicloud.bucket.get,
                        'bucket.put': scicloud.bucket.put,
                        'bucket.iterlist': scicloud.bucket.iterlist,
                        'bucket.list': scicloud.bucket.list,
                        'bucket.info': scicloud.bucket.info,
                        'bucket.remove': scicloud.bucket.remove,
                        'bucket.remove-prefix': scicloud.bucket.remove_prefix,
                        'bucket.get-md5': scicloud.bucket.get_md5,
                        'bucket.sync-from-scicloud': scicloud.bucket.sync_from_scicloud,
                        'bucket.sync-to-scicloud': scicloud.bucket.sync_to_scicloud,
                        'bucket.make-public' : functions.bucket_make_public,
                        'bucket.make-private' : scicloud.bucket.make_private,
                        'bucket.is-public' : scicloud.bucket.is_public,
                        'bucket.public-url-folder' : scicloud.bucket.public_url_folder,
                        'bucket.mpsafe-get' : scicloud.bucket.mpsafe_get,
                        'realtime.request': scicloud.realtime.request,
                        'realtime.release': scicloud.realtime.release,
                        'realtime.list': scicloud.realtime.list,
                        'volume.list': scicloud.volume.get_list,
                        'volume.create': scicloud.volume.create,
                        'volume.mkdir': scicloud.volume.mkdir,
                        'volume.sync': scicloud.volume.sync,
                        'volume.delete': scicloud.volume.delete,
                        'volume.ls': scicloud.volume.ls,
                        'volume.rm': scicloud.volume.rm,
                        'env.list': scicloud.environment.list_envs,
                        'env.list-bases': scicloud.environment.list_bases,
                        'env.create': scicloud.environment.create,
                        'env.edit-info': scicloud.environment.edit_info,
                        'env.clone': scicloud.environment.clone,
                        'env.modify': scicloud.environment.modify,
                        'env.get-hostname': scicloud.environment.get_setup_hostname,
                        'env.get-keypath': scicloud.environment.get_key_path,
                        'env.save': scicloud.environment.save,
                        'env.shutdown': scicloud.environment.shutdown,
                        'env.save-shutdown': scicloud.environment.save_shutdown,
                        'env.delete': scicloud.environment.delete,
                        'env.ssh': scicloud.environment.ssh,
                        'env.rsync': scicloud.environment.rsync,
                        'env.run-script': scicloud.environment.run_script,
                        #'queue.list': scicloud.queue.list,
                        #'queue.create': scicloud.queue.create,
                        #'queue.delete': scicloud.queue.delete,
                        'cron.register': functions.cron_register, # move to rest?
                        'cron.deregister' : scicloud.cron.deregister,
                        'cron.list' : scicloud.cron.list,
                        'cron.run' : scicloud.cron.manual_run,
                        'cron.info' : scicloud.cron.info,
                        'wait-for.status' : scicloud.wait_for.status,
                        'wait-for.port' : scicloud.wait_for.port,
                        }
    
    # maps the called function to another function for printing the output
    printer_mapping = {'status' : key_val_printer('jid', 'status'),
                       'info' : scicloud_info_printer,
                       'result' : scicloud_result_printer,
                       'ssh-info' : dict_printer(['address', 'port', 'username', 'identity']),
                       'rest.list' : list_printer('label'),
                       'rest.info' : dict_printer(['label', 'uri', 'signature', 'output_encoding', 'description']),
                       'realtime.request': dict_printer(['request_id', 'type', 'cores', 'start_time']),
                       'realtime.list': list_of_dicts_printer(['request_id', 'type', 'cores', 'start_time']),
                       'files.list': list_printer('filename'),
                       'bucket.list': list_printer('filename'),
                       'bucket.iterlist': list_printer('filename'),
                       'bucket.info': bucket_info_printer,
                       'volume.list': list_of_dicts_printer(['name', 'mnt_path', 'created', 'desc']),
                       'volume.ls': volume_ls_printer,
                       'env.list': list_of_dicts_printer(['name', 'status', 'action', 'created', 'last_modified']),
                       'env.list-bases': list_of_dicts_printer(['name', 'distro', 'python_version']),
                       'env.get-hostname': no_newline_printer,
                       'env.get-keypath': no_newline_printer,
                       'env.ssh': no_newline_printer,
                       'env.run-script': no_newline_printer,
                       'queue.list' : list_printer('label'),
                       'cron.list' : list_printer('label'),
                       'cron.info' : dict_printer(['label', 'schedule', 'last_run', 'last_jid', 'created', 'creator_host', 'func_name']),
                       'wait-for.port' : dict_printer(['address', 'port']),
                       }
    
    json_printer_mapping = {'result' : scicloud_result_json_printer}

    try:
        # execute function
        ret = function_mapping[function_name](**kwargs) 

        if parsed_args._output == 'json':
            # ordereddict issue.. fix once we on 2.7 only
            if (isinstance(ret, dict) or isinstance(ret, DictMixin)) and type(ret) != dict: 
                ret = dict(ret)
            json_printer = json_printer_mapping.get(function_name)
            if json_printer:
                json_printer(ret, parsed_args._output != 'no-header', kwargs)
            else:
                print json.dumps(ret)
        else:
            if function_name in printer_mapping:
                # use a printer_mapping if it exists
                # this is how dict/tables and lists with columns are printed
                printer_mapping[function_name](ret, parsed_args._output != 'no-header', kwargs)
            else:
                if isinstance(ret, (tuple, list)):
                    # if it's just  list with no mapping, just print it
                    for item in ret:
                        print item
                elif ret or isinstance(ret, bool):
                    # cases where the output is just a string or number or bool
                    print ret
                else:
                    # if the output is None, print nothing
                    pass
            
    except scicloud.CloudException, e:
        if parsed_args._output == 'json':
            sys.stderr.write( json.dumps(e.args ) )
        else:
            # error thrown by scicloud client library)
            sys.stderr.write(str(e)+'\n')
            sys.exit(3)
        
    except Exception, e:
        # unexpected errors
        sys.stderr.write('Got unexpected error\n')
        traceback.print_exc(file = sys.stderr)
        sys.exit(1)
        
    else:
        sys.exit(0)



"""
Defines basic connection object

Copyright (c) 2009 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@piscicloud.com

The scicloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see 
http://www.gnu.org/licenses/lgpl-2.1.html
"""

from ..scicloud import CloudException

class CloudConnection(object):
    """Abstract connection class to deal with low-level communication of scicloud adapter"""
    
    _isopen = False
    _adapter = None
    
    @property
    def opened(self): 
        """Returns whether the connection is open"""
        return self._isopen
    
    def open(self):
        """called when this connection is to be used"""
        if self._adapter and not self._adapter.opened:
            self._adapter.open()
        self._isopen = True
    
    def close(self):
        """called when this connection is no longer needed"""
        if not self.opened:
            raise CloudException("%s: Cannot close a closed connection", str(self))
        self._isopen = False
        
    @property
    def adapter(self):
        return self._adapter
    
    def needs_restart(self, **kwargs):
        """Called to determine if the scicloud must be restarted due to different connection parameters"""
        return False
    
    def job_add(self, params, logdata = None):
        raise NotImplementedError
    
    def jobs_join(self, jids, timeout = None):
        """
        Allows connection to manage joining
        If connection manages joining, it should return a list of statuses  
        describing the finished job
        Else, return False
        """
        return False
    
    def jobs_map(self, params, mapargs, mapkwargs = None, logdata = None):
        raise NotImplementedError
    
    def jobs_result(self, jids):
        raise NotImplementedError
    
    def jobs_kill(self, jids):
        raise NotImplementedError

    def jobs_delete(self, jids):
        raise NotImplementedError
    
    def jobs_info(self, jids, info_requested):
        raise NotImplementedError
    
    def is_simulated(self):        
        raise NotImplementedError
    
    def connection_info(self):
        return {'opened': self.opened, 'connection_type' :None}
    
    def modules_check(self, modules):
        pass
    
    def modules_add(self, modules):
        pass
    
    def packages_list(self):
        """
        Get list of packages from server
        """
        return []

    def force_adapter_report(self):
        """
        Should the SerializationReport for the SerializationAdapter be coerced to be instantiated?
        """
        return False
    
    def report_name(self):
        raise NotImplementedError

    def get_report_dir(self):
        raise TypeError('get_report_dir is only valid on connection hooks')


"""
Functions for setting up the local machine's API Keys.
This module is only intended for internal use
"""
"""
Copyright (c) 2012 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@piscicloud.com

The scicloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see 
http://www.gnu.org/licenses/lgpl-2.1.html    
"""

import os
import sys
import getpass
import time

import scicloud
from scicloud.util import credentials, fix_sudo_path
import webbrowser
import random
import BaseHTTPServer

import logging
scicloudLog = logging.getLogger('Cloud.setup_machine')

WEB_AUTH_PATH = 'accounts/request_setup_token/'
WEB_TIMEOUT = 600
CALLBACK_PATH = '/auth_callback/'
SUCCESS_REDIRECT = 'accounts/received_setup_token/'

# base url to connect to PiCloud
web_base_url = None

def _is_browser_graphical(browser_obj):
    """Return if a given webbrowser.Browser object refers to a graphical browser"""
    if isinstance(browser_obj, webbrowser.UnixBrowser):
        return browser_obj.background
    if isinstance(browser_obj, webbrowser.BackgroundBrowser):
        return True
    always_graphical = ['Konqueror', 'WindowsDefault', 'MacOSX', 'MacOSXOSAScript']
    return browser_obj.__class__.__name__ in always_graphical
    

def graphical_web_open(url, new=1, autoraise=True):
    """Similar to webbrowser.open, but only displays graphical webbrowsers.
    Must use protected interfaces due to lack of an API to access such information
    
    Returns True if web was opened; False otherwise"""
    
    for name in webbrowser._tryorder:
        browser = webbrowser.get(name)
        
        if _is_browser_graphical(browser) and browser.open(url, new, autoraise):
            return True
    return False    

def system_browser_may_be_graphical():
    """Return true if there may be a graphical webbrowser on this machine"""
    if sys.platform == 'darwin' or sys.platform[:3] == 'win':
        return True
    return "DISPLAY" in os.environ

class PiCloudHTTPHandler(BaseHTTPServer.BaseHTTPRequestHandler):
    """HTTP Server handler to retrieve email and token from PiCloud webserver
    
    Webserver issues a 302 redirect to localhost
    Credentials are inside get parameters
    """ 
    
    def log_message(self, format, *args):
        """Don't write to stderr"""
        scicloudLog.debug(format, *args)
    
    def do_GET(self):
        global web_base_url
        
        # Parse out email and temporary authorization key
        valid = 0
        path_tuple = self.path.split('?')
        if len(path_tuple) == 2 and path_tuple[0] == CALLBACK_PATH:
            get_vars = path_tuple[1]
            all_gets = get_vars.split('&')
            for getbind in all_gets:
                get_tuple = getbind.split('=')
                if len(get_tuple) == 2:
                    if get_tuple[0] == 'token':
                        self.server.auth_token = get_tuple[1]
                        scicloudLog.debug('Received token from server')
                        valid += 1
                if get_tuple[0] == 'email':
                    self.server.email = get_tuple[1]
                    scicloudLog.debug('Received email from server: %s' % self.server.email)
                    valid += 1                
                    
        if  valid == 2: # redirect to web page telling user to close
            self.send_response(302)
            self.send_header('Location',web_base_url + SUCCESS_REDIRECT)
            self.end_headers()
        else:
            self.send_error(404, 'Invalid path')

def start_http_server():
    """Returned spawn server object"""    
    for port in xrange(30000+random.randint(0,500), 31000):
        # create listener socket, bind it to the port, and start listening
        try:
            server_address = ('localhost', port)
            httpd = BaseHTTPServer.HTTPServer(server_address, PiCloudHTTPHandler)
            # configure variables response handler writes to
            httpd.email = None
            httpd.auth_token = None            
            
            httpd.timeout = WEB_TIMEOUT
            scicloudLog.debug('Listening server started on port %s' % httpd.server_port)
            return httpd
        except Exception, e:
            if e[0] == 98:   # error code corresponds to socket in use
                scicloudLog.debug('socket %s already in use', port)
                continue
            scicloudLog.error('failed to listen on port %s', port)
            raise    

def web_acquire_token():
    """Acquire the secure token via the browser"""
    global web_base_url
    try:
        server = start_http_server()
    except Exception, e:
        print 'Could not start local webserver due to %s!' % e,
        return
    
    api_url = scicloud._getscicloudnetconnection().url
    web_base_url = api_url.replace('//api.', '//') # map to web url from api    
    #web_base_url = 'http://localhost:8000/' # temp! 
    
    full_url = web_base_url + WEB_AUTH_PATH + '?redirect_port=%s' % server.server_port 
    if not graphical_web_open(full_url):
        print 'Could not launch webbrowser!',
        return
    
    # TOOD: Catch timeout
    start_time = time.time()
    while not server.auth_token:
        if time.time() > start_time + WEB_TIMEOUT:
            print 'Did not receive credentials within %s seconds.' % WEB_TIMEOUT
            break
                    
        server.handle_request()
    server.server_close()
    return server.email, server.auth_token
    

def setup_machine(email=None, password=None, api_key=None):
    """Prompts user for login information and then sets up api key on the
    local machine
    
    If api_key is a False value, interpretation is:
        None: create api key
        False: Prompt for key selection
    
    """
    
    # Disable simulator -- we need to initiate net connections
    scicloud.config.use_simulator = False
    scicloud.config.commit()
    
    auth_token = None # authentication key derived from webserver injection
    
    # connect to scivm
    scicloud._getscicloud().open()
    
    interactive_mode = not (email and password)
    
    if system_browser_may_be_graphical() and interactive_mode:
        print 'To authenticate your computer, a web browser will be launched.  Please login if prompted and follow instructions on screen.\n'
        raw_input('Press ENTER to continue')
        
        
        auth_info = web_acquire_token()
        if not auth_info:
            print 'Reverting to email/password authentication.\n'
        else:
            email, password = auth_info
            print '\n'
    
    if interactive_mode:
        print 'Please enter your PiCloud account login information.\nIf you do not have an account, please create one at http://www.scivm.com\n' + \
        'Note that a password is required. If you have not set one, set one at https://www.scivm.com/accounts/settings/\n'
    
    try:
        if email:            
            print 'Setup will proceed using this E-mail: %s' % email
        else:
            email = raw_input('E-mail: ')
            
        if not password:
            password = getpass.getpass('Password: ')
        
        if not api_key:   
            
            if interactive_mode:                
                keys = scicloud.account.list_keys(email, password, active_only=True)
                
                print """\nPiCloud uses API Keys, rather than your login information, to authenticate
    your machine. In the event your machine is compromised, you can deactivate
    your API Key to disable access. In this next step, you can choose to use
    an existing API Key for this machine, or create a new one. We recommend that
    each machine have its own unique key."""
                
                print '\nYour API Key(s)'
                for key in keys:
                    print key
                
                api_key = raw_input('\nPlease select an API Key or just press enter to create a new one automatically: ')
                if api_key:
                    key = scicloud.account.get_key(email, password, api_key)
                else:
                    key = scicloud.account.create_key(email, password)
                    print 'API Key: %s' % key['api_key']
            else:
                api_key = scicloud.config.api_key
                if api_key and api_key != 'None':
                    print 'Using existing API Key: %s' % api_key
                    key = {'api_key' : api_key}
                else:
                    key = scicloud.account.create_key(email, password)
                    print 'API Key: %s' % key['api_key']
            
        else:
            key = scicloud.account.get_key(email, password, api_key)
            print 'API Key: %s' % key['api_key']
        
        # save all key credentials
        if 'api_secretkey' in key:
            credentials.save_keydef(key)
        
        # set config and write it to file
        scicloud.config.api_key = key['api_key']
        scicloud.config.commit()
        scicloud.scicloudconfig.flush_config()        
                
        
        # if user is running "scivm setup" with sudo, we need to chown
        # the config file so that it's owned by user and not root.
        fix_sudo_path(os.path.join(scicloud.scicloudconfig.fullconfigpath,scicloud.scicloudconfig.configname))
        fix_sudo_path(scicloud.scicloudconfig.fullconfigpath)
        
        try:
            import platform
            conn = scicloud._getscicloudnetconnection()
            conn.send_request('report/install/', {'hostname': platform.node(),
                                                  'language_version': platform.python_version(),
                                                  'language_implementation': platform.python_implementation(),
                                                  'platform': platform.platform(),
                                                  'architecture': platform.machine(),
                                                  'processor': platform.processor(),
                                                  'pyexe_build' : platform.architecture()[0]
                                                  })
        except:
            pass
        
    except EOFError:
        sys.stderr.write('Got EOF. Please run "scivm setup" to complete installation.\n')
        sys.exit(1)
    except KeyboardInterrupt:
        sys.stderr.write('Got Keyboard Interrupt. Please run "scivm setup" to complete installation.\n')
        sys.exit(1)
    except scicloud.CloudException, e:
        sys.stderr.write(str(e)+'\n')
        sys.exit(3)
    else:
        print '\nSetup successful!'



"""
This module holds convenience functions for accessing ssh information 

Copyright (c) 2012 `PiCloud, Inc. <http://www.picloud.com>`_.  All rights reserved.

email: contact@piscicloud.com

The scicloud package is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This package is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this package; if not, see 
http://www.gnu.org/licenses/lgpl-2.1.html
"""

from .. import _getscicloud, _getscicloudnetconnection

def get_ssh_info(jid, timeout=None):
    """Return dictionary describing how to access job's ssh port.
    
    Blocks until ssh service is ready. An optional ``timeout`` in seconds may
    be specified. Returned dictionary provides:
    * address: Address to connect to
    * port: Port to connect to
    * username: username to connect with
    * identity: Path of identity file to use 
    
    """
    from . import get_connection_info
    from ..util.credentials import get_sshkey_path
    
    conn_dct = get_connection_info(jid, 22, timeout=timeout)

    api_key = _getscicloudnetconnection().api_key
    key_path = get_sshkey_path(api_key)
    conn_dct['identity'] = key_path
    return conn_dct
